// Source file for the GAPS mesh analysis program



////////////////////////////////////////////////////////////////////////
// Include files 
////////////////////////////////////////////////////////////////////////

#include "R3Shapes/R3Shapes.h"



////////////////////////////////////////////////////////////////////////
// Program arguments
////////////////////////////////////////////////////////////////////////

static char *input_mesh_name = NULL;
static char *input_points_name = NULL;
static char *input_importances_name = NULL;
static char *input_descriptors_name = NULL;
static char *output_irsa_name = NULL;
static int print_verbose = 0;
static int print_debug = 0;



////////////////////////////////////////////////////////////////////////
// Data variables
////////////////////////////////////////////////////////////////////////

static R3Mesh *mesh = NULL;
static R3MeshVertex **points = NULL;
static RNScalar *importances = NULL;
static RNScalar **distances = NULL;
static RNScalar **descriptors = NULL;
static RNScalar **similarities = NULL;
static RNScalar **balances = NULL;
static R3MeshProperty *irsa = NULL;
static int npoints = 128;
static int ndescriptors = 32;
static RNScalar target_interpoint_distance = 0.5;
static RNScalar balance_exponent = 1;
static RNScalar distance_sigma = 0.05;
static RNScalar similarity_sigma = 1;



////////////////////////////////////////////////////////////////////////
// Mesh input function
////////////////////////////////////////////////////////////////////////

static int
ReadMesh(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate mesh
  mesh = new R3Mesh();
  assert(mesh);

  // Read mesh from file
  if (!mesh->ReadFile(filename)) {
    fprintf(stderr, "Unable to read mesh from %s\n", filename);
    return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read mesh from %s ...\n", filename);
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Faces = %d\n", mesh->NFaces());
    printf("  # Edges = %d\n", mesh->NEdges());
    printf("  # Vertices = %d\n", mesh->NVertices());
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Point input/creation functions
////////////////////////////////////////////////////////////////////////

static int
ReadPoints(const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = fopen(filename, "r");
  if (!fp) {
    fprintf(stderr, "Unable to open points file: %s\n", filename);
    return 0;
  }

  // Read vertices from file (vertexid is first column of each row)
  RNArray<R3MeshVertex *> vertices;
  char buffer[2048];
  while (fgets(buffer, 2048, fp)) {
    int vertex_index;
    if (sscanf(buffer, "%d", &vertex_index) == 1) {
      // Check vertex
      if ((vertex_index < 0) || (vertex_index >= mesh->NVertices())) {
        fprintf(stderr, "Invalid vertex index %d in %s\n", vertex_index, filename);
        return 0;
      }

      // Find vertex in mesh
      R3MeshVertex *vertex = mesh->Vertex(vertex_index);
      if (!vertex) {
        fprintf(stderr, "Unable to find vertex %d in %s\n", vertex_index, filename);
        return 0;
      }

      // Add vertex to array
      vertices.Insert(vertex);
    }
  }

  // Close file
  fclose(fp);

  // Copy into global variables
  npoints = vertices.NEntries();
  points = new R3MeshVertex * [ npoints ];
  for (int i = 0; i < npoints; i++) points[i] = vertices[i];

  // Print statistics
  if (print_verbose) {
    printf("Read points from %s ...\n", filename);
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Points = %d\n", npoints);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
CreatePoints(void) 
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Check/adjust the number of points
  if (npoints > mesh->NVertices()) npoints = mesh->NVertices();

  // Allocate array of points (vertices)
  points = new R3MeshVertex * [ npoints ];
  if (!points) {
    fprintf(stderr, "Unable to allocate array of s\n");
    return 0;
  }

  // Select random starting vertex
  RNArray<R3MeshVertex *> vertices;
  int i = (int) (RNRandomScalar() * mesh->NVertices());
  R3MeshVertex *vertex = mesh->Vertex(i);
  vertices.Insert(vertex);

  // Iteratively select furthest vertex
  for (int i = 0; i < npoints; i++) {

    // Compute distances from selected vertices to all vertices
    RNScalar *d = mesh->DijkstraDistances(vertices);

    // Find furthest vertex
    RNScalar furthest_distance = 0;
    R3MeshVertex *furthest_vertex = NULL;
    for (int j = 0; j < mesh->NVertices(); j++) {
      if (d[j] > furthest_distance) {
        furthest_distance = d[j];
        furthest_vertex = mesh->Vertex(j);
      }
    }

    // Check if found furthest vertex
    if (!furthest_vertex) break;

    // Add furthest vertex to selected set
    if (i == 0) vertices.Truncate(0);
    vertices.Insert(furthest_vertex);

    // Delete distances
    delete [] d;
  }

  // Copy into return values
  points = new R3MeshVertex * [ npoints ];
  for (int i = 0; i < npoints; i++) points[i] = vertices[i];

  // Print statistics
  if (print_verbose) {
    printf("Created points ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Points = %d\n", npoints);
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Importance creation functions
////////////////////////////////////////////////////////////////////////

static int
ReadImportances(const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate importances
  importances = new RNScalar [ npoints ];
  if (!importances) {
    fprintf(stderr, "Unable to create array of importances\n");
    return 0;
  }

  // Read property file
  R3MeshPropertySet propertyset(mesh);
  if (!propertyset.Read(input_importances_name)) {
    fprintf(stderr, "Unable to read importances: %s\n", filename);
    return 0;
  }

  // Check importance file
  if (propertyset.NProperties() == 0) {
    fprintf(stderr, "Importance file has not properties: %s\n", filename);
    return 0;
  }

  // Compute importance values
  RNScalar maximum = 0;
  R3MeshProperty *property = propertyset.Property(0);
  for (int i = 0; i < npoints; i++) {
    int vertex_id = mesh->VertexID(points[i]);
    importances[i] = property->VertexValue(vertex_id);
    if (importances[i] > maximum) maximum = importances[i];
  }

  // Normalize importance values
  if (maximum > 0) {
    for (int i = 0; i < npoints; i++) {
      importances[i] /= maximum;
    }
  }

  // Check importance values
  if (maximum == 0) {
    for (int i = 0; i < npoints; i++) {
      importances[i] = 1;
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Read importances ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Importances = %d\n", npoints);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
CreateImportances(void)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate importances
  importances = new RNScalar [ npoints ];
  if (!importances) {
    fprintf(stderr, "Unable to create array of importances\n");
    return 0;
  }

  // Compute importances
  for (int i = 0; i < npoints; i++) {
    importances[i] = 1;
  }

  // Print statistics
  if (print_verbose) {
    printf("Created importances ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Importances = %d\n", npoints);
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Distance creation functions
////////////////////////////////////////////////////////////////////////

static int
CreateDistances(void)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate distances
  distances = new RNScalar * [ npoints ];

  // Compute distances
  for (int i = 0; i < npoints; i++) {
    distances[i] = mesh->DijkstraDistances(points[i]);
  }

  // Normalize distances
  RNArea area = mesh->Area();
  RNScalar normalization_factor = (area > 0) ? 1.0 / sqrt(mesh->Area()) : 1;
  for (int i = 0; i < npoints; i++) {
    for (int j = 0; j < mesh->NVertices(); j++) {
      distances[i][j] *= normalization_factor;
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Created dijkstra distances ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Points = %d\n", npoints);
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Descriptor input/creation functions
////////////////////////////////////////////////////////////////////////

static RNScalar **
ReadDescriptors(const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Read descriptors from file
  R3MeshPropertySet propertyset(mesh);
  if (!propertyset.Read(filename)) {
    fprintf(stderr, "Unable to read descriptors from %s\n", filename);
    return 0;
  }
    
  // Allocate descriptors
  descriptors = new RNScalar * [ npoints ];
  if (!descriptors) {
    fprintf(stderr, "Unable to allocate descriptors\n");
    return 0;
  }

  // Copy descriptors for points
  ndescriptors = propertyset.NProperties();
  for (int i = 0; i < npoints; i++) {
    R3MeshVertex *point = points[i];
    descriptors[i] = new RNScalar [ ndescriptors  ];
    for (int j = 0; j < ndescriptors; j++) {
      R3MeshProperty *property = propertyset.Property(j);
      descriptors[i][j] = property->VertexValue(mesh->VertexID(point)) ;
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Read descriptors from %s ...\n", filename);
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Descriptors = %d\n", ndescriptors);
    fflush(stdout);
  }

  // Return descriptors
  return descriptors;
}



static int
CreateDescriptors(void)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate memory for descriptors
  descriptors = new RNScalar * [ npoints ];
  for (int i = 0; i < npoints; i++) {
    descriptors[i] = new RNScalar [ ndescriptors ];
    for (int j = 0; j < ndescriptors; j++) {
      descriptors[i][j] = 0;
    }
  }

  // Compute normalization factor
  RNScalar vote = 1.0 / mesh->NVertices();
  RNScalar normalization = ndescriptors / 3.0;

  // Compute histogram of distances
  for (int i = 0; i < npoints; i++) {
    for (int j = 0; j < mesh->NVertices(); j++) {
      RNScalar bin = normalization * distances[i][j];
      int bin1 = (int) bin;
      int bin2 = bin1 + 1;
      RNScalar t = bin - bin1;
      if (bin1 >= ndescriptors) bin1 = ndescriptors-1;
      if (bin2 >= ndescriptors) bin2 = ndescriptors-1;
      descriptors[i][bin1] += (1-t) * vote;
      descriptors[i][bin2] += t * vote;
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Created descriptors with histograms of dijkstra distances ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Descriptors = %d\n", ndescriptors);
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Similarity functions
////////////////////////////////////////////////////////////////////////

static int
CreateSimilarities(void)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate memory for similarities
  similarities = new RNScalar * [ npoints ];
  for (int i = 0; i < npoints; i++) {
    similarities[i] = new RNScalar [ npoints ];
    for (int j = 0; j < npoints; j++) {
      similarities[i][j] = 1;
    }
  }

#if 1
  // Get convenient variables
  RNScalar squared_similarity_sigma = similarity_sigma * similarity_sigma;
  if (squared_similarity_sigma == 0) squared_similarity_sigma = RN_EPSILON;

  // Allocate temporary array of property differences
  RNScalar *squared_diffs = new RNScalar [ npoints * npoints ];
  for (int i = 0; i < npoints*npoints; i++) squared_diffs[i] = 0;

  // Compute squared differences between descriptors
  int count = 0;
  for (int i = 0; i < npoints; i++) {
    for (int j = i+1; j < npoints; j++) {
      // Compute squared difference between descriptors
      RNScalar squared_diff = 0;
      for (int k = 0; k < ndescriptors; k++) {
        RNScalar diff = descriptors[i][k] - descriptors[j][k];
        squared_diff += diff * diff;
      }

      // Remember squared differences
      squared_diffs[count++] = squared_diff;
      similarities[i][j] = squared_diff;
      similarities[j][i] = squared_diff;
    }
  }

  // Sort squared diffs
  qsort(squared_diffs, count, sizeof(RNScalar), RNCompareScalars);
  RNScalar minimum_squared_diff = squared_diffs [ 0 ];
  RNScalar cutoff_squared_diff = squared_diffs [ npoints ];
  if (cutoff_squared_diff <= minimum_squared_diff) cutoff_squared_diff = minimum_squared_diff + RN_EPSILON;

  // Normalize similarities
  for (int i = 0; i < npoints; i++) {
    for (int j = i+1; j < npoints; j++) {
      RNScalar normalized_squared_diff = (similarities[i][j] - minimum_squared_diff) / (cutoff_squared_diff - minimum_squared_diff);
      similarities[i][j] = exp ( -normalized_squared_diff / squared_similarity_sigma);
      similarities[j][i] = similarities[i][j];
    }
  }

  // Delete squared differences
  delete [] squared_diffs;
#else
  // Allocate temporary array of property differences
  RNScalar *l2diffs = new RNScalar [ npoints ];

  // Compute L2 differences between descriptors
  for (int i = 0; i < npoints; i++) {
    // Compute differences between descriptors
    for (int j = 0; j < npoints; j++) {
      l2diffs[j] = 0;
      for (int k = 0; k < ndescriptors; k++) {
        RNScalar diff = descriptors[i][k] - descriptors[j][k];
        l2diffs[j] += diff * diff;
      }
    }

    // Compute similarities
    for (int j = 0; j < npoints; j++) {
      // Find best remaining l2diff
      int best_index = -1;
      RNScalar best_value = FLT_MAX;
      for (int k = 0; k < npoints; k++) {
        if ((l2diffs[k] >= 0) && (l2diffs[k] < best_value)) {
          best_value = l2diffs[k];
          best_index = k;
        }
      }

      // Assign similarity
      if (best_index >= 0) {
        similarities[i][best_index] = 1.0 / (j+1);
        l2diffs[best_index] = -1;
      }
    }
  }

  // Delete temporary array
  delete [] l2diffs;
#endif

  // Print statistics
  if (print_verbose) {
    // Compute statistics
    int count = 0;
    RNScalar total_similarity = 0;
    RNScalar minimum_similarity = FLT_MAX;
    RNScalar maximum_similarity = -FLT_MAX;
    for (int i = 0; i < npoints; i++) {
      for (int j = i+1; j < npoints; j++) {
        if (similarities[i][j] < minimum_similarity) minimum_similarity = similarities[i][j];
        if (similarities[i][j] > maximum_similarity) maximum_similarity = similarities[i][j];
        total_similarity += similarities[i][j];
        count++;
      }
    }

    // Print stuff
    printf("Created similarities ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Similarities = %d\n", count);
    printf("  Minimum = %g\n", minimum_similarity);
    printf("  Maximum = %g\n", maximum_similarity);
    printf("  Mean = %g\n", total_similarity / count);
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Similarity functions
////////////////////////////////////////////////////////////////////////

static int
CreateBalances(void)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  int count = 0;

  // Allocate memory for balances
  balances = new RNScalar * [ npoints ];
  for (int i = 0; i < npoints; i++) {
    balances[i] = new RNScalar [ npoints ];
    for (int j = 0; j < npoints; j++) {
      balances[i][j] = 0;
    }
  }

  // Allocate temporary memory for vertex areas
  RNArea *areas = new RNArea [ mesh->NVertices() ];
  for (int k = 0; k < mesh->NVertices(); k++) {
    R3MeshVertex *vertex = mesh->Vertex(k);
    areas[k] = mesh->VertexArea(vertex);
  }

  // Compute balances between areas on sides of split
  for (int i = 0; i < npoints; i++) {
    for (int j = i+1; j < npoints; j++) {
      // Count area on sides of split
      RNArea area0 = 0, area1 = 0;
      for (int k = 0; k < mesh->NVertices(); k++) {
        if (distances[i][k] < distances[j][k]) area0 += areas[k];
        else if (distances[i][k] > distances[j][k]) area1 += areas[k];
      }
      
      // Compute balance
      if ((area0 > 0) && (area1 > 0)) {
        RNScalar balance = (area0 < area1) ? area0 / area1 : area1 / area0;
        balances[i][j] = balance;
        balances[j][i] = balance;
        count++;
      }
    }
  }

  // Normalize balances 
  for (int i = 0; i < npoints; i++) {
    for (int j = i+1; j < npoints; j++) {
      balances[i][j] = pow(balances[i][j], balance_exponent);
    }
  }

  // Delete termporary memory
  delete [] areas;

  // Print statistics
  if (print_verbose) {
    printf("Created area balances ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Balances = %d\n", count);
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// IRSA computation
////////////////////////////////////////////////////////////////////////

static int
CreateIRSA(void)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  int nvotes = 0;

  // Get convenient variables
  RNScalar squared_target_interpoint_distance = target_interpoint_distance *  target_interpoint_distance;
  RNScalar distance_sigma_squared = distance_sigma * distance_sigma;
  RNScalar min_vote = 0.001;

  // Allocate property
  irsa = new R3MeshProperty(mesh, "IRSA");
  if (!irsa) {
    fprintf(stderr, "Unable to allocate IRSA\n");
    return 0;
  }

  // Consider all pairs of sample points
  for (int ia = 0; ia < npoints; ia++) {
    // R3MeshVertex *a = points[ia];
    for (int ib = ia+1; ib < npoints; ib++) {
      R3MeshVertex *b = points[ib];
      RNScalar ab_vote = 1;

      // Consider importance of a
      ab_vote *= importances[ia];
      if (ab_vote < min_vote) continue;

      // Consider importance of b
      ab_vote *= importances[ib];
      if (ab_vote < min_vote) continue;

      // Consider distance ab
      RNScalar dab = distances[ia][mesh->VertexID(b)];
      RNScalar dab_delta = dab - target_interpoint_distance;
      ab_vote *= exp(-dab_delta * dab_delta / squared_target_interpoint_distance);
      if (ab_vote < min_vote) continue;

      // Consider similarity of a and b
      ab_vote *= similarities[ia][ib];
      if (ab_vote < min_vote) continue;

      // Consider balance of a and b
      ab_vote *= balances[ia][ib];
      if (ab_vote < min_vote) continue;

      if (print_debug) printf("%d %d : %g %g %g : %g\n", ia, ib, 
        similarities[ia][ib], balances[ia][ib], distances[ia][mesh->VertexID(b)], ab_vote);

      // Consider all pairs of sample points
      for (int ip = 0; ip < npoints; ip++) {
        if ((ip == ia) || (ip == ib)) continue;
        R3MeshVertex *p = points[ip];
        for (int iq = ip+1; iq < npoints; iq++) {
          if ((iq == ia) || (iq == ib)) continue;
          R3MeshVertex *q = points[iq];
          RNScalar pq_vote = ab_vote;

          // Consider importance of p
          pq_vote *= importances[ip];
          if (pq_vote < min_vote) continue;

          // Consider importance of q
          pq_vote *= importances[iq];
          if (pq_vote < min_vote) continue;

          // Consider distance pq
          RNScalar dpq = distances[ip][mesh->VertexID(q)];
          RNScalar dpq_delta = dpq - target_interpoint_distance;
          pq_vote *= exp(-dpq_delta * dpq_delta / squared_target_interpoint_distance);
          if (pq_vote < min_vote) continue;

          // Consider similarity of p and q
          pq_vote *= similarities[ip][iq];
          if (pq_vote < min_vote) continue;

          // Consider balance of p and q
          pq_vote *= balances[ip][iq];
          if (pq_vote < min_vote) continue;

          // if (print_debug) printf("  %d %d %g\n", ip, iq, pq_vote);

          // Check distances ap vs. bq
          RNScalar dap = distances[ia][mesh->VertexID(p)];
          RNScalar dbq = distances[ib][mesh->VertexID(q)];
          RNScalar dap_dbq = dap - dbq;
          pq_vote *= exp(-dap_dbq * dap_dbq / distance_sigma_squared);
          if (pq_vote < min_vote) continue;

          // Check distances aq vs. bp
          RNScalar daq = distances[ia][mesh->VertexID(q)];
          RNScalar dbp = distances[ib][mesh->VertexID(p)];
          RNScalar daq_dbp = daq - dbp;
          pq_vote *= exp(-daq_dbp * daq_dbp / distance_sigma_squared);
          if (pq_vote < min_vote) continue;

          // Cast vote
          for (int ic = 0; ic < mesh->NVertices(); ic++) {
            R3MeshVertex *c = mesh->Vertex(ic);
            RNScalar vote = pq_vote;

            // Check distances ac vs. bc
            RNScalar dac = distances[ia][mesh->VertexID(c)];
            RNScalar dbc = distances[ib][mesh->VertexID(c)];
            RNScalar dac_dbc = dac - dbc;
            vote *= exp(-dac_dbc * dac_dbc / distance_sigma_squared);
            if (vote < min_vote) continue;

            // Check distances pc vs. qc
            RNScalar dpc = distances[ip][mesh->VertexID(c)];
            RNScalar dqc = distances[iq][mesh->VertexID(c)];
            RNScalar dpc_dqc = dpc - dqc;
            vote *= exp(-dpc_dqc * dpc_dqc / distance_sigma_squared);
            if (vote < min_vote) continue;

            // if (print_debug) printf("    %g\n", vote);

            // Add vote value
            irsa->AddVertexValue(ic, vote);

            // Cast votes
            nvotes++;
          }
        }
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Computed IRSA property ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Votes = %d\n", nvotes);
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Output function
////////////////////////////////////////////////////////////////////////

static int
WriteProperty(R3MeshProperty *property, char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write property to file
  if (!property->Write(filename)) {
    fprintf(stderr, "Unable to write property to %s\n", filename);
    return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("Wrote property to %s ...\n", filename);
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Program argument parsing
////////////////////////////////////////////////////////////////////////

static int 
ParseArgs(int argc, char **argv)
{
  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = 1;
      else if (!strcmp(*argv, "-debug")) print_debug = 1;
      else if (!strcmp(*argv, "-npoints")) { argv++; argc--; npoints = atoi(*argv); }
      else if (!strcmp(*argv, "-points")) { argv++; argc--; input_points_name = *argv; }
      else if (!strcmp(*argv, "-importances")) { argv++; argc--; input_importances_name = *argv; }
      else if (!strcmp(*argv, "-descriptors")) { argv++; argc--; input_descriptors_name = *argv; }
      else if (!strcmp(*argv, "-distance_sigma")) { argv++; argc--; distance_sigma = atof(*argv); }
      else if (!strcmp(*argv, "-similarity_sigma")) { argv++; argc--; similarity_sigma = atof(*argv); }
      else if (!strcmp(*argv, "-balance_exponent")) { argv++; argc--; balance_exponent = atof(*argv); }
      else if (!strcmp(*argv, "-target_interpoint_distance")) { argv++; argc--; target_interpoint_distance = atof(*argv); }
      else { fprintf(stderr, "Invalid program argument: %s", *argv); exit(1); }
      argv++; argc--;
    }
    else {
      if (!input_mesh_name) input_mesh_name = *argv;
      else if (!output_irsa_name) output_irsa_name = *argv;
      else { fprintf(stderr, "Invalid program argument: %s", *argv); exit(1); }
      argv++; argc--;
    }
  }

  // Check input filename
  if (!input_mesh_name || !output_irsa_name) {
    fprintf(stderr, "Usage: irsa meshfile propertyfile [options]\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Main program
////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(1);

  // Read mesh file
  if (!ReadMesh(input_mesh_name)) exit(-1);

  // Read/compute points
  if (input_points_name) { if (!ReadPoints(input_points_name)) exit(-1); }
  else { if (!CreatePoints()) exit(-1); }
    
  // Read/compute point importances
  if (input_importances_name) { if (!ReadImportances(input_importances_name)) exit(-1); }
  else { if (!CreateImportances()) exit(-1); }
    
  // Compute distances
  if (!CreateDistances()) exit(-1); 

  // Read/compute descriptors
  if (input_descriptors_name) { if (!ReadDescriptors(input_descriptors_name)) exit(-1); }
  else { if (!CreateDescriptors()) exit(-1); }

  // Compute similarities
  if (!CreateSimilarities()) exit(-1); 

  // Compute balances
  if (!CreateBalances()) exit(-1); 

  // Compute IRSA
  if (!CreateIRSA()) exit(-1);
    
  // Write IRSA file
  if (!WriteProperty(irsa, output_irsa_name)) exit(-1);

  // Return success 
  return 0;
}

















