// Source file for the plane reflection program



////////////////////////////////////////////////////////////////////////
// Include files 
////////////////////////////////////////////////////////////////////////

#include "R3Shapes/R3Shapes.h"



////////////////////////////////////////////////////////////////////////
// Global variables
////////////////////////////////////////////////////////////////////////

// File names
static char *input_mesh_name = NULL;
static char *input_points_name = NULL;
static char *input_properties_name = NULL;
static char *input_grid_name = NULL;
static char *input_planes_name = NULL;
static char *output_points_name = NULL;
static char *output_point_properties_name = NULL;
static char *output_vertex_properties_name = NULL;
static char *output_grid_name = NULL;
static char *output_planes_name = NULL;

// Parameters
static R3Vector in_plane_vector = R3unknown_vector;
static double radius = 0.5;
static double min_value = 0;
static double min_weight = RN_EPSILON;
static int max_points = 4096;
static int max_angles = 2;

// Statistics
static int plane_count = 0;

// Verbosity
static int print_verbose = 0;
static int print_debug = 0;



////////////////////////////////////////////////////////////////////////
// Type definitions
////////////////////////////////////////////////////////////////////////

struct Descriptor {
public:
  Descriptor(void) 
    : values(NULL), nvalues(0) {};
  Descriptor(const Descriptor& descriptor) : values(NULL), nvalues(descriptor.nvalues)  
    { values = new RNScalar [ nvalues ]; for (int i = 0; i < nvalues; i++) values[i] = descriptor.values[i];  };
  ~Descriptor(void) 
    { if (values) delete [] values; };
public:
  RNScalar *values;
  int nvalues;
};

struct Point {
public:
  Point(R3MeshFace *face = NULL, const R3Point& position = R3zero_point, const R3Vector& normal = R3zero_vector) 
    : face(face), position(position), normal(normal), descriptor(),  
      best_plane(NULL), best_value(0), best_radial_value(0), best_proximity_value(0), best_mirror_value(0),   
      pointset(NULL), pointset_index(-1) {};
  Point(const Point& point) 
    : face(point.face), position(point.position), normal(point.normal), descriptor(point.descriptor), 
      best_plane(point.best_plane), best_value(point.best_value), best_radial_value(point.best_radial_value),
      best_proximity_value(point.best_proximity_value), best_mirror_value(point.best_mirror_value), 
      pointset(NULL), pointset_index(-1) {};
  ~Point(void)  
    {};
public:
  R3MeshFace *face;
  R3Point position;
  R3Vector normal;
  Descriptor descriptor;
  struct Plane *best_plane;
  RNScalar best_value;
  RNScalar best_radial_value;
  RNScalar best_proximity_value;
  RNScalar best_mirror_value;
  struct PointSet *pointset;  
  int pointset_index;
};

struct PointSet {
public:
  PointSet(R3Mesh *mesh) 
    : mesh(mesh), kdtree(NULL), points() {};
  ~PointSet(void)
    { if (kdtree) delete kdtree; };
  int NPoints(void) const 
    { return points.NEntries(); };
  Point *KthPoint(int k) const 
    { return points.Kth(k); };
  void InsertPoint(Point *point) 
    { if (kdtree) { delete kdtree; kdtree = NULL; }
      point->pointset = this;
      point->pointset_index = points.NEntries();
      points.Insert(point); };
  Point *FindClosest(const R3Point& query, RNLength max_distance) 
    { static Point tmp; 
      static int position_offset = (unsigned char *) &(tmp.position) - (unsigned char *) &tmp;
      if (!kdtree) kdtree = new R3Kdtree<Point *>(points, position_offset);
      return kdtree->FindClosest(query, 0, max_distance); };
  int FindAll(const R3Point& query, RNLength max_distance, RNArray<Point *>& result) 
    { static Point tmp; 
      static int position_offset = (unsigned char *) &(tmp.position) - (unsigned char *) &tmp;
      if (!kdtree) kdtree = new R3Kdtree<Point *>(points, position_offset);
      return kdtree->FindAll(query, 0, max_distance, result); };
public:
  R3Mesh *mesh;
  R3Kdtree<Point *> *kdtree;
  RNArray<Point *> points;
};

struct Plane {
public:
  Plane(void)
    : position(RN_UNKNOWN, RN_UNKNOWN, RN_UNKNOWN), normal(RN_UNKNOWN, RN_UNKNOWN, RN_UNKNOWN), radius(RN_UNKNOWN),
      value(0), radial_value(0), proximity_value(0), mirror_value(0), weight(0) {};
  Plane(const R3Point& position, const R3Vector& normal, RNLength radius, 
     RNScalar value, RNScalar radial_value, RNScalar proximity_value, RNScalar mirror_value, RNScalar weight)
    :  position(position), normal(normal), radius(radius), 
       value(value), radial_value(radial_value), proximity_value(proximity_value), mirror_value(mirror_value), weight(weight) {};
  ~Plane(void)  
    {};
public:
  R3Point position;
  R3Vector normal;
  RNScalar radius;
  RNScalar value;
  RNScalar radial_value;
  RNScalar proximity_value;
  RNScalar mirror_value;
  RNScalar weight;
};

struct PlaneSet {
public:
  PlaneSet(R3Mesh *mesh) 
    : mesh(mesh), planes() {};
  ~PlaneSet(void)
    {};
  int NPlanes(void) const 
    { return planes.NEntries(); };
  Plane *KthPlane(int k) const 
    { return planes.Kth(k); };
  void InsertPlane(Plane *plane) 
    { planes.Insert(plane); };
public:
  R3Mesh *mesh;
  RNArray<Plane *> planes;
};



////////////////////////////////////////////////////////////////////////
// Mesh input functions
////////////////////////////////////////////////////////////////////////

static R3Mesh *
ReadMesh(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  if (print_verbose) {
    printf("Reading mesh from %s ...\n", filename);
    fflush(stdout);
  }

  // Allocate mesh
  R3Mesh *mesh = new R3Mesh();
  if (!mesh) {
    fprintf(stderr, "Unable to allocate mesh.\n");
    return NULL;
  }

  // Read mesh from file
  if (!mesh->ReadFile(filename)) {
    fprintf(stderr, "Unable to read mesh from %s\n", filename);
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Faces = %d\n", mesh->NFaces());
    printf("  # Edges = %d\n", mesh->NEdges());
    printf("  # Vertices = %d\n", mesh->NVertices());
    fflush(stdout);
  }

  // Return mesh
  return mesh;
}



////////////////////////////////////////////////////////////////////////
// Point creation/input/output functions
////////////////////////////////////////////////////////////////////////

static PointSet *
CreatePoints(R3Mesh *mesh, int max_points)
{ 
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Create point set
  PointSet *pointset = new PointSet(mesh);
  if (!pointset) {
    fprintf(stderr, "Unable to allocate point set\n");
    return NULL;
  }

  // Allocate temporary data
  RNArea *cumulative_area = new RNArea [ mesh->NFaces() ];

  // Count cumulative area of faces
  RNArea total_area = 0;
  for (int i = 0; i < mesh->NFaces(); i++) {
    R3MeshFace *face = mesh->Face(i);
    RNArea face_area = mesh->FaceArea(face);
    total_area += face_area;
    cumulative_area[i] = total_area;;
  }
    
  // Generate points
  RNSeedRandomScalar();
  while (pointset->NPoints() < max_points) {
    // Generate a random number
    RNScalar r = RNRandomScalar() * total_area;

    // Find face ID
    int min_face_id = 0;
    int max_face_id = mesh->NFaces()-1;
    int face_id = (min_face_id + max_face_id) / 2;
    while (min_face_id < max_face_id) {
      if (min_face_id + 1 >= max_face_id) { 
        if (r < cumulative_area[min_face_id]) { 
          face_id = min_face_id; 
          break; 
        }
        else { 
          face_id = max_face_id; 
          break; 
        }
      }
      else {
        if (cumulative_area[face_id] < r) {
          min_face_id = face_id + 1;
          face_id = (min_face_id + max_face_id) / 2;
        }
        else if (cumulative_area[face_id] > r) {
          max_face_id = face_id;
          face_id = (min_face_id + max_face_id) / 2;
        }
        else {
          break;
        }
      }
    }

    // Find point on face
    R3MeshFace *face = mesh->Face(face_id);
    R3MeshVertex *v0 = mesh->VertexOnFace(face, 0);
    R3MeshVertex *v1 = mesh->VertexOnFace(face, 1);
    R3MeshVertex *v2 = mesh->VertexOnFace(face, 2);
    const R3Point& p0 = mesh->VertexPosition(v0);
    const R3Point& p1 = mesh->VertexPosition(v1);
    const R3Point& p2 = mesh->VertexPosition(v2);
    const R3Vector& n0 = mesh->VertexNormal(v0);
    const R3Vector& n1 = mesh->VertexNormal(v1);
    const R3Vector& n2 = mesh->VertexNormal(v2);
    RNScalar r1 = sqrt(RNRandomScalar());
    RNScalar r2 = RNRandomScalar();
    RNScalar t0 = (1.0 - r1);
    RNScalar t1 = r1 * (1.0 - r2);
    RNScalar t2 = r1 * r2;
    R3Point position = t0*p0 + t1*p1 + t2*p2;
    R3Vector normal = t0*n0 + t1*n1 + t2*n2;

    // Create point
    Point *point = new Point(face, position, normal);
    pointset->InsertPoint(point);
  }

  // Delete temporary data
  delete [] cumulative_area;

  // Print statistics
  if (print_verbose) {
    printf("Created points ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Points = %d\n", pointset->NPoints());
    fflush(stdout);
  }

  // Return point set
  return pointset;
}  



static PointSet *
ReadPoints(R3Mesh *mesh, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  if (print_verbose) {
    fprintf(stdout, "Reading points from %s ...\n", filename);
    fflush(stdout);
  }

  // Open file
  FILE *fp = fopen(filename, "r");
  if (!fp) {
    fprintf(stderr, "Unable to open points file: %s\n", filename);
    return 0;
  }

  // Create point set
  PointSet *pointset = new PointSet(mesh);
  if (!pointset) {
    fprintf(stderr, "Unable to allocate point set\n");
    return NULL;
  }

  // Read points from file 
  char buffer[2048];
  while (fgets(buffer, 2048, fp)) {
    int face_index;
    double barycentrics[3];
    if (sscanf(buffer, "%d%lf%lf%lf", &face_index, &barycentrics[0], &barycentrics[1], &barycentrics[2]) == 4) {
      // Check face
      if ((face_index < 0) || (face_index >= mesh->NFaces())) {
        fprintf(stderr, "Invalid face index %d in %s\n", face_index, filename);
        return 0;
      }

      // Find face in mesh
      R3MeshFace *face = mesh->Face(face_index);
      if (!face) {
        fprintf(stderr, "Unable to find face %d in %s\n", face_index, filename);
        return 0;
      }

      // Find point position
      R3Point position = mesh->FacePoint(face, barycentrics);

      // Find point normal
      R3Vector normal = mesh->FaceNormal(face);

      // Create point
      Point *point = new Point();
      point->face = face;
      point->position = position;
      point->normal = normal;

      // Insert point
      pointset->InsertPoint(point);
    }
  }

  // Close file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Points = %d\n", pointset->NPoints());
    fflush(stdout);
  }

  // Return points
  return pointset;
}



static int
WritePoints(PointSet *pointset, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  if (print_verbose) {
    fprintf(stdout, "Writing points to %s ...\n", filename);
    fflush(stdout);
  }

  // Open file
  FILE *fp = fopen(filename, "w");
  if (!fp) {
    fprintf(stderr, "Unable to open output point file %s\n", filename);
    return 0;
  }

  // Write points
  for (int i = 0; i < pointset->NPoints(); i++) {
    Point *point = pointset->KthPoint(i);
    const R3Point& position = point->position;
    fprintf(fp, "%g %g %g\n", position.X(), position.Y(), position.Z());
  }

  // Close file
  fclose(fp);

  // Print message
  if (print_verbose) {
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Points = %d\n", pointset->NPoints());
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Grid creation/input/output functions
////////////////////////////////////////////////////////////////////////

static R3Grid *
CreateGrid(PointSet *pointset, RNLength radius)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  if (print_verbose) {
    printf("Creating grid ...\n");
    fflush(stdout);
  }

  // Get bounding box
  R3Mesh *mesh = pointset->mesh;
  const R3Box& bbox = mesh->BBox();

  // Compute grid parameters
  int min_res = 3;
  int max_res = 256;
  RNScalar spacing = 0.05 * radius;
  int xres = (int) (bbox.XLength() / spacing + 0.5);
  int yres = (int) (bbox.YLength() / spacing + 0.5);
  int zres = (int) (bbox.ZLength() / spacing + 0.5);
  xres = (xres > min_res) ? ((xres < max_res) ? xres : max_res) : min_res;
  yres = (yres > min_res) ? ((yres < max_res) ? yres : max_res) : min_res;
  zres = (zres > min_res) ? ((zres < max_res) ? zres : max_res) : min_res;

  // Allocate grid
  R3Grid *grid = new R3Grid(xres, yres, zres, bbox);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    return NULL;
  }

#if 0
  // Fill grid
  for (int i = 0; i < mesh->NFaces(); i++) {
    R3MeshFace *face = mesh->Face(i);
    R3MeshVertex *v0 = mesh->VertexOnFace(face, 0);
    R3MeshVertex *v1 = mesh->VertexOnFace(face, 1);
    R3MeshVertex *v2 = mesh->VertexOnFace(face, 2);
    const R3Point& p0 = mesh->VertexPosition(v0);
    const R3Point& p1 = mesh->VertexPosition(v1);
    const R3Point& p2 = mesh->VertexPosition(v2);
    grid->RasterizeWorldTriangle(p0, p1, p2, 1.0);
  }

  // Process grid 
  grid->Threshold(0.5, 0.0, 1.0);
  grid->Blur(0.25 * radius / spacing);
#else
  // Convenient variables 
  RNScalar dr_sigma = 0.5 * radius;
  RNScalar dr_scale = -1.0 / (2.0 * dr_sigma * dr_sigma);
  RNScalar da_sigma = RN_PI / 16.0;
  RNScalar da_scale = -1.0 / (2.0 * da_sigma * da_sigma);

  // Fill grid
  for (int i = 0; i < pointset->NPoints(); i++) {
    Point *point0 = pointset->KthPoint(i);
    const R3Point& position0 = point0->position;
    const R3Vector& normal0 = point0->normal;
    for (int j = i+1; j < pointset->NPoints(); j++) {
      Point *point1 = pointset->KthPoint(j);
      const R3Point& position1 = point1->position;
      const R3Vector& normal1 = point1->normal;

      // Start with full weight
      RNScalar weight = 1.0;

      // Consider distance between points
      RNLength r = 0.5 * R3Distance(position0, position1);
      RNLength dr = fabs(r - radius);
      weight *= exp(dr_scale * dr);
      
      // Consider angle between normals
      RNScalar dot = fabs(normal0.Dot(normal1));
      RNScalar da = (dot < 1.0) ? acos(dot) : 0;
      weight *= exp(da_scale * da);

      // Cast vote for midpoint
      R3Point midpoint = 0.5 * (position0 + position1);
      grid->RasterizeWorldPoint(midpoint.X(), midpoint.Y(), midpoint.Z(), weight);
    }
  }

  // Process grid
  grid->Blur(1);
#endif

  // Print statistics
  if (print_verbose) {
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", grid->L1Norm());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



static R3Grid *
ReadGrid(char *input_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  if (print_verbose) {
    printf("Reading grid ...\n");
    fflush(stdout);
  }

  // Allocate a grid
  R3Grid *grid = new R3Grid();
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    return NULL;
  }

  // Read grid
  if (!grid->ReadFile(input_name)) {
    fprintf(stderr, "Unable to read grid file %s\n", input_name);
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", grid->L1Norm());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



static int 
WriteGrid(R3Grid *grid, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  if (print_verbose) {
    printf("Writing grid to %s ...\n", filename);
    fflush(stdout);
  }

  // Write grid
  if (!grid->WriteFile(filename)) return 0;

  // Print statistics
  if (print_verbose) {
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", grid->L1Norm());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Plane creation/input/output functions
////////////////////////////////////////////////////////////////////////

static Plane *
CreatePlane(PlaneSet *planeset, const RNArray<Point *>& neighborhood_points, R3MeshSearchTree *kdtree, 
  const R3Point& position, const R3Vector& normal, RNScalar radius)
{
  // Get convenient variables
  RNLength neighborhood_sigma = radius;
  RNLength correspondence_sigma = 0.1 * neighborhood_sigma;
  RNLength correspondence_radius = 3 * correspondence_sigma;
  RNScalar neighborhood_factor = -1.0 / (2.0 * neighborhood_sigma * neighborhood_sigma);
  RNScalar correspondence_factor = -1.0 / (2.0 * correspondence_sigma * correspondence_sigma);

  // Create reflecton plane
  R3Vector reflection_normal = normal; 
  if (in_plane_vector != R3unknown_vector) {
    reflection_normal.Cross(in_plane_vector);
    reflection_normal.Cross(in_plane_vector);
  }    
  RNLength reflection_normal_length = reflection_normal.Length();
  if (RNIsZero(reflection_normal_length)) return NULL;
  reflection_normal /= reflection_normal_length;
  R3Plane reflection_plane(position, reflection_normal);

  // Update statistics
  plane_count++;

  // Initialize data
  RNScalar plane_weight = 0;
  RNScalar plane_value = 0;
  RNScalar plane_radial_value = 0;
  RNScalar plane_proximity_value = 0;
  RNScalar plane_mirror_value = 0;
  RNScalar *point_radial_values = new RNScalar [ neighborhood_points.NEntries() ];
  RNScalar *point_proximity_values = new RNScalar [ neighborhood_points.NEntries() ];
  RNScalar *point_mirror_values = new RNScalar [ neighborhood_points.NEntries() ];
  for (int i = 0; i < neighborhood_points.NEntries(); i++) {
    point_radial_values[i] = 0;
    point_proximity_values[i] = 0;
    point_mirror_values[i] = 0;
  }

  // Compute support for plane amongst points in neighborhood
  for (int i = 0; i < neighborhood_points.NEntries(); i++) {
    Point *point0 = neighborhood_points.Kth(i);

    // Get position of point0 mirrored across plane
    R3Point mirror0 = point0->position;
    mirror0.Mirror(reflection_plane);

    // Find point1 closest to point0 after mirroring across plane
    R3MeshIntersection closest;
    kdtree->FindClosest(mirror0, closest, 0, correspondence_radius);
    if (closest.type == R3_MESH_NULL_TYPE) continue;
    
    // Compute weight
    RNLength squared_neighborhood_distance = R3SquaredDistance(point0->position, position);
    RNScalar weight = exp(neighborhood_factor * squared_neighborhood_distance);
    plane_weight += weight;

    // Compute radial value
    RNScalar radial_value = weight;
    point_radial_values[i] = radial_value;
    plane_radial_value += radial_value;
        
    // Compute proximity value
    RNScalar proximity = R3Distance(reflection_plane, point0->position);
    RNScalar proximity_value = exp(neighborhood_factor * proximity * proximity);
    plane_proximity_value += weight * proximity_value;
    point_proximity_values[i] = proximity_value;
    
    // Compute mirror value
    RNLength correspondence_squared_distance = R3SquaredDistance(mirror0, closest.point);
    RNScalar mirror_value = exp(correspondence_factor * correspondence_squared_distance);
    plane_mirror_value += weight * mirror_value;
    point_mirror_values[i] = mirror_value;

    // Compute value
    plane_value += weight * mirror_value;
  }

  // Check plane weight
  if ((plane_weight < min_weight) || (plane_weight <= 0)) {
    delete [] point_radial_values; 
    delete [] point_proximity_values; 
    delete [] point_mirror_values; 
    return NULL; 
  }

  // Divide values by total radial weight (do not favor planes with many or more nearby points)
  plane_value /= plane_weight;
  plane_proximity_value /= plane_weight;
  plane_mirror_value /= plane_weight;
  plane_radial_value /= neighborhood_points.NEntries();
  if ((plane_value < min_value) || (plane_value <= 0)) {
    delete [] point_radial_values; 
    delete [] point_proximity_values; 
    delete [] point_mirror_values; 
    return NULL; 
  }

  // Create plane
  Plane *plane = new Plane(position, reflection_normal, radius, 
    plane_value, plane_radial_value, plane_proximity_value, plane_mirror_value, plane_weight);
  if (!plane) {
    fprintf(stderr, "Unable to allocate plane\n");
    delete [] point_radial_values; 
    delete [] point_proximity_values; 
    delete [] point_mirror_values; 
    return NULL;
  }

  // Update points
  for (int i = 0; i < neighborhood_points.NEntries(); i++) {
    Point *point = neighborhood_points.Kth(i);

    // Compute point value stuff
    RNScalar point_radial_value = point_radial_values[i];
    RNScalar point_proximity_value = point_proximity_values[i];
    RNScalar point_mirror_value = point_mirror_values[i];
    RNScalar point_value = point_proximity_value * point_mirror_value * plane_value;

    // Remember best point value
    if (point_value > point->best_value) {
      point->best_value = point_value;
      point->best_radial_value = point_radial_value;
      point->best_proximity_value = point_proximity_value;
      point->best_mirror_value = point_mirror_value;
      point->best_plane = plane;
    }
  }

  // Insert plane
  planeset->InsertPlane(plane);

  // Delete data
  delete [] point_mirror_values;
  delete [] point_proximity_values;
  delete [] point_radial_values;

  // Print message
  if (print_debug) {
    printf("P   %g %g %g   %g %g %g   %g   %g %g %g %g %g\n", 
      plane->position.X(), plane->position.Y(), plane->position.Z(), 
      plane->normal.X(), plane->normal.Y(), plane->normal.Z(), plane->radius, 
      plane->value, plane->radial_value, plane->proximity_value, plane->mirror_value, plane->weight);
  }

  // Return plane
  return plane;
}



static int
CreatePlane(PlaneSet *planeset, PointSet *pointset, R3MeshSearchTree *kdtree, 
  const R3Point& position, const R3Vector& normal, RNScalar radius)
{
  // Find points in neighborhood
  RNArray<Point *> neighborhood_points;
  const int min_neighborhood_points = 5;
  RNLength neighborhood_sigma = radius;
  RNLength neighborhood_radius = 3 * neighborhood_sigma;
  if (!pointset->FindAll(position, neighborhood_radius, neighborhood_points)) return 1;
  if (neighborhood_points.NEntries() < min_neighborhood_points) return 1;

  // Create plane
  CreatePlane(planeset, neighborhood_points, kdtree, position, normal, radius);
  
  // Return success
  return 1;
}



static int
CreatePlanes(PlaneSet *planeset, PointSet *pointset, R3MeshSearchTree *kdtree, 
  const R3Point& position, RNScalar radius)
{
  // Find points in neighborhood
  RNArray<Point *> neighborhood_points;
  const int min_neighborhood_points = 5;
  RNLength neighborhood_sigma = radius;
  RNLength neighborhood_radius = 3 * neighborhood_sigma;
  if (!pointset->FindAll(position, neighborhood_radius, neighborhood_points)) return 1;
  if (neighborhood_points.NEntries() < min_neighborhood_points) return 1;

  // Create array of R3Points (for PCA)
  RNArray<R3Point *> positions;
  for (int i = 0; i < neighborhood_points.NEntries(); i++) {
    Point *point = neighborhood_points.Kth(i);
    positions.Insert(&(point->position));
  }

  // Create PCA planes
  RNScalar variances[3] = { 0.0, 0.0, 0.0 };
  R3Point centroid = R3Centroid(positions);
  R3Triad axes = R3PrincipleAxes(centroid, positions, NULL, variances);
  for (int i = 0; i < 3; i++) {
    if (variances[i] > 0.01 * radius * radius) {
      CreatePlane(planeset, neighborhood_points, kdtree, position, axes[i], radius);
    }
  }

  // Consider regularly spaced planes
  if (in_plane_vector != R3unknown_vector) {
    // Consider normals by rotating around in_plane_vector
    int dim = in_plane_vector.MinDimension();
    R3Vector vector0 = R3xyz_triad.Axis(dim) % in_plane_vector;
    vector0.Normalize();
    for (int j = 0; j < max_angles; j++) {
      R3Vector vector = vector0;
      vector.Rotate(in_plane_vector, j * RN_PI / max_angles); 
      vector.Normalize();
      CreatePlane(planeset, neighborhood_points, kdtree, position, vector, radius);
    }
  }
  else {
    // Consider normals by sampling pitch and yaw
    CreatePlane(planeset, pointset, kdtree, position, R3posz_vector, radius);
    for (int i = 0; i < max_angles; i++) {
      RNAngle pitch = i * RN_PI / max_angles; 
      for (int j = 0; j < max_angles; j++) {
        RNAngle yaw = j * RN_PI / max_angles; 
        R3Vector vector(pitch, yaw);
        CreatePlane(planeset, neighborhood_points, kdtree, position, vector, radius);
      }
    }
  }
  
  // Return success
  return 1;
}



static int
CreateLocalPlanes(PlaneSet *planeset, PointSet *pointset, R3Grid *grid, R3MeshSearchTree *kdtree, RNScalar radius)
{
  // Find properties of grid
  R3Grid maxima(*grid); 
  maxima.MaskNonMaxima(4);
  maxima.Threshold(0.01, 0.0, R3_GRID_KEEP_VALUE);
  R3Grid dx(*grid); dx.Gradient(RN_X);
  R3Grid dy(*grid); dy.Gradient(RN_Y);
  R3Grid dz(*grid); dz.Gradient(RN_Z);

  // Consider planes at local maxima of grid oriented along gradient
  for (int iz = 0; iz < maxima.ZResolution(); iz++) {
    for (int iy = 0; iy < maxima.YResolution(); iy++) {
      for (int ix = 0; ix < maxima.XResolution(); ix++) {
        RNScalar value = maxima.GridValue(ix, iy, iz);
        if (value == R2_GRID_UNKNOWN_VALUE) continue;
        if (value <= 0) continue;
        R3Point position = maxima.WorldPosition(ix, iy, iz);

        // Create planes for gradient direction
        RNCoord nx = dx.GridValue(ix, iy, iz);
        if (nx == R2_GRID_UNKNOWN_VALUE) continue;
        RNCoord ny = dy.GridValue(ix, iy, iz);
        if (ny == R2_GRID_UNKNOWN_VALUE) continue;
        RNCoord nz = dz.GridValue(ix, iy, iz);
        if (nz == R2_GRID_UNKNOWN_VALUE) continue;
        R3Vector normal(nx, ny, nz);
        CreatePlane(planeset, pointset, kdtree, position, normal, radius);
        if (in_plane_vector != R3unknown_vector) {
          R3Vector vector = normal % in_plane_vector;
          CreatePlane(planeset, pointset, kdtree, position, vector, radius);
        }

        // Create planes for a variety of other directions 
        CreatePlanes(planeset, pointset, kdtree, position, radius);
      }
    }
  }

  // Return success
  return 1;
}



static int
CreateGlobalPlanes(PlaneSet *planeset, PointSet *pointset, R3MeshSearchTree *kdtree, RNScalar radius)
{
  // Get convenient variables
  R3Mesh *mesh = pointset->mesh;
  R3Point centroid = mesh->Centroid();
  R3Triad pca_axes = mesh->PrincipleAxes(&centroid);

  // Create planes for a variety of directions through centroid
  CreatePlanes(planeset, pointset, kdtree, centroid, radius);

  // Create planes through global centroid along principle axes
  for (int j = 0; j < 3; j++) {
    R3Vector normal = pca_axes.Axis(j);
    if (RNIsZero(normal.Length())) continue;
    CreatePlane(planeset, pointset, kdtree, centroid, normal, radius);
  }

  // Return success
  return 1;
}



static PlaneSet *
CreatePlanes(PointSet *pointset, R3Grid *grid, RNScalar radius)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  if (print_verbose) {
    printf("Creating planes ...\n");
    fflush(stdout);
  }

  // Create plane set
  R3Mesh *mesh = pointset->mesh;
  PlaneSet *planeset = new PlaneSet(mesh);
  if (!planeset) {
    fprintf(stderr, "Unable to allocate plane set\n");
    return NULL;
  }

  // Create mesh kdtree
  R3MeshSearchTree *kdtree = new R3MeshSearchTree(mesh);
  if (!kdtree) {
    fprintf(stderr, "Unable to create mesh kdtree\n");
    return NULL;
  }

  // Create global planes
  if (!CreateGlobalPlanes(planeset, pointset, kdtree, radius)) {
    delete planeset;
    delete kdtree;
    return NULL;
  }

  // Create local planes
  if (!CreateLocalPlanes(planeset, pointset, grid, kdtree, radius)) {
    delete planeset;
    delete kdtree;
    return NULL;
  }

  // Delete mesh kdtree
  delete kdtree;

  // Print statistics
  if (print_verbose) {
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Planes = %d\n", planeset->NPlanes());
    printf("  # Planes Checked = %d\n", plane_count);
    fflush(stdout);
  }

  // Return planes
  return planeset;
}



static PlaneSet *
ReadPlanes(R3Mesh *mesh, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  if (print_verbose) {
    printf("Reading planes from %s ...\n", filename);
    fflush(stdout);
  }

  // Open file
  FILE *fp = fopen(filename, "r");
  if (!fp) {
    fprintf(stderr, "Unable to open planes file: %s\n", filename);
    return 0;
  }

  // Create plane set
  PlaneSet *planeset = new PlaneSet(mesh);
  if (!planeset) {
    fprintf(stderr, "Unable to allocate plane set\n");
    return NULL;
  }

  // Read planes from file 
  char buffer[2048];
  while (fgets(buffer, 2048, fp)) {
    double px, py, pz, nx, ny, nz, r, value, radial_value, proximity_value, mirror_value, weight;
    if (sscanf(buffer, "%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf", &px, &py, &pz, &nx, &ny, &nz, &r, 
      &value, &radial_value, &proximity_value, &mirror_value, &weight) == (unsigned int) 12) {
      // Create plane
      Plane *plane = new Plane();
      plane->position = R3Point(px, py, pz);
      plane->normal = R3Vector(nx, ny, nz);
      plane->radius = r;
      plane->value = value;
      plane->radial_value = radial_value;
      plane->proximity_value = proximity_value;
      plane->mirror_value = mirror_value;
      plane->weight = weight;

      // Insert plane
      planeset->InsertPlane(plane);
    }
  }

  // Close file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Planes = %d\n", planeset->NPlanes());
    fflush(stdout);
  }

  // Return plane set
  return planeset;
}



static int
WritePlanes(PlaneSet *planeset, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  if (print_verbose) {
    fprintf(stdout, "Writing planes to %s ...\n", filename);
    fflush(stdout);
  }

  // Open file
  FILE *fp = fopen(filename, "w");
  if (!fp) {
    fprintf(stderr, "Unable to open output plane file %s\n", filename);
    return 0;
  }

  // Write planes
  for (int i = 0; i < planeset->NPlanes(); i++) {
    Plane *plane = planeset->KthPlane(i);
    fprintf(fp, "%g %g %g   %g %g %g   %g   %g %g %g %g %g\n", 
      plane->position.X(), plane->position.Y(), plane->position.Z(), 
      plane->normal.X(), plane->normal.Y(), plane->normal.Z(), plane->radius, 
      plane->value, plane->radial_value, plane->proximity_value, plane->mirror_value, plane->weight);
  }

  // Close file
  fclose(fp);

  // Print message
  if (print_verbose) {
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Planes = %d\n", planeset->NPlanes());
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Plane processing functions
////////////////////////////////////////////////////////////////////////

static int 
ComparePlanes(const void *data1, const void *data2)
{
  Plane *plane1 = *((Plane **) data1);
  Plane *plane2 = *((Plane **) data2);
  if (plane1->value > plane2->value) return -1;
  else if (plane1->value < plane2->value) return 1;
  else return 0;
}



static Plane * 
FindPlane(PlaneSet *planeset, Plane *query, 
  RNLength position_distance_tolerance, RNAngle normal_angle_tolerance)
{
  // Search for equivalent plane
  RNScalar min_dot = cos(normal_angle_tolerance);
  RNScalar max_dd = position_distance_tolerance * position_distance_tolerance;
  for (int i = 0; i < planeset->NPlanes(); i++) {
    Plane *plane = planeset->KthPlane(i);
    RNLength dd = R3SquaredDistance(query->position, plane->position);
    if (dd > max_dd) continue;
    RNScalar dot = query->normal.Dot(plane->normal);
    if (fabs(dot) < min_dot) continue;
    return plane;
  }

  // Not found
  return NULL;
}



static int
SortPlanes(PlaneSet *planeset)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  if (print_verbose) {
    fprintf(stdout, "Sorting planes ...\n");
    fflush(stdout);
  }

  // Sort planes
  planeset->planes.Sort(ComparePlanes);

  // Print message
  if (print_verbose) {
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Planes = %d\n", planeset->NPlanes());
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
UniquifyPlanes(PlaneSet *planeset)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  if (print_verbose) {
    fprintf(stdout, "Uniquifying planes ...\n");
    fflush(stdout);
  }

  // Copy planes and then empty
  RNArray<Plane *> planes = planeset->planes;
  planeset->planes.Empty();

  // Re-insert unique planes
  for (int i = 0; i < planes.NEntries(); i++) {
    Plane *plane = planes.Kth(i);
    if (FindPlane(planeset, plane, radius, 0.5 * RN_PI / max_angles)) delete plane;
    else planeset->InsertPlane(plane);
  }

  // Print message
  if (print_verbose) {
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Planes = %d\n", planeset->NPlanes());
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Property creation/input/output functions
////////////////////////////////////////////////////////////////////////

static int
CreateProperties(PointSet *pointset)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  int nproperties = 0;
  if (print_verbose) {
    printf("Creating properties ...\n");
    fflush(stdout);
  }

  // Print statistics
  if (print_verbose) {
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Points = %d\n", pointset->NPoints());
    printf("  # Properties = %d\n", nproperties);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
ReadProperties(PointSet *pointset, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  int nproperties = 0;

  // Print statistics
  if (print_verbose) {
    printf("Read properties from %s ...\n", filename);
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Points = %d\n", pointset->NPoints());
    printf("  # Properties = %d\n", nproperties);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
WritePointProperties(PointSet *pointset, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = fopen(filename, "w");
  if (!fp) {
    fprintf(stderr, "Unable to open output point properties file %s\n", filename);
    return 0;
  }

  // Write header
  fprintf(fp, "@relation PlaneReflection\n");

  // Write attribute names
  fprintf(fp, "@attribute PointValue real\n");
  fprintf(fp, "@attribute PointRadialValue real\n");
  fprintf(fp, "@attribute PointProximityValue real\n");
  fprintf(fp, "@attribute PointMirrorValue real\n");
  fprintf(fp, "@attribute PlaneValue real\n");
  fprintf(fp, "@attribute PlaneRadialValue real\n");
  fprintf(fp, "@attribute PlaneProximityValue real\n");
  fprintf(fp, "@attribute PlaneMirrorValue real\n");
  fprintf(fp, "@attribute PlaneWeight real\n");

  // Write data
  fprintf(fp, "@data\n");
  for (int i = 0; i < pointset->NPoints(); i++) {
    Point *point = pointset->KthPoint(i);
    Plane *plane = point->best_plane;
    fprintf(fp, "%g ", (point) ? point->best_value : 0.0);
    fprintf(fp, "%g ", (point) ? point->best_radial_value : 0.0);
    fprintf(fp, "%g ", (point) ? point->best_proximity_value : 0.0);
    fprintf(fp, "%g ", (point) ? point->best_mirror_value : 0.0);
    fprintf(fp, "%g ", (plane) ? plane->value : 0.0);
    fprintf(fp, "%g ", (plane) ? plane->radial_value : 0.0);
    fprintf(fp, "%g ", (plane) ? plane->proximity_value : 0.0);
    fprintf(fp, "%g ", (plane) ? plane->mirror_value : 0.0);
    fprintf(fp, "%g ", (plane) ? plane->weight : 0.0);
    fprintf(fp, "\n");
  }

  // Close file
  fclose(fp);

  // Print message
  if (print_verbose) {
    fprintf(stdout, "Wrote point properties to %s ...\n", filename);
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Points = %d\n", pointset->NPoints());
    printf("  # Properties = 4\n");
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
WriteVertexProperties(PointSet *pointset, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = fopen(filename, "w");
  if (!fp) {
    fprintf(stderr, "Unable to open output point properties file %s\n", filename);
    return 0;
  }

  // Write header
  fprintf(fp, "@relation PlaneReflection\n");

  // Write attribute names
  fprintf(fp, "@attribute PointValue real\n");
  fprintf(fp, "@attribute PointRadialValue real\n");
  fprintf(fp, "@attribute PointProximityValue real\n");
  fprintf(fp, "@attribute PointMirrorValue real\n");
  fprintf(fp, "@attribute PlaneValue real\n");
  fprintf(fp, "@attribute PlaneRadialValue real\n");
  fprintf(fp, "@attribute PlaneProximityValue real\n");
  fprintf(fp, "@attribute PlaneMirrorValue real\n");
  fprintf(fp, "@attribute PlaneWeight real\n");

  // Write data
  fprintf(fp, "@data\n");
  R3Mesh *mesh = pointset->mesh;
  for (int i = 0; i < mesh->NVertices(); i++) {
    R3MeshVertex *vertex = mesh->Vertex(i);
    const R3Point& position = mesh->VertexPosition(vertex);
    Point *point = pointset->FindClosest(position, radius);
    Plane *plane = (point) ? point->best_plane : NULL;
    fprintf(fp, "%g ", (point) ? point->best_value : 0.0);
    fprintf(fp, "%g ", (point) ? point->best_radial_value : 0.0);
    fprintf(fp, "%g ", (point) ? point->best_proximity_value : 0.0);
    fprintf(fp, "%g ", (point) ? point->best_mirror_value : 0.0);
    fprintf(fp, "%g ", (plane) ? plane->value : 0.0);
    fprintf(fp, "%g ", (plane) ? plane->radial_value : 0.0);
    fprintf(fp, "%g ", (plane) ? plane->proximity_value : 0.0);
    fprintf(fp, "%g ", (plane) ? plane->mirror_value : 0.0);
    fprintf(fp, "%g ", (plane) ? plane->weight : 0.0);
    fprintf(fp, "\n");
  }

  // Close file
  fclose(fp);

  // Print message
  if (print_verbose) {
    fprintf(stdout, "Wrote point properties to %s ...\n", filename);
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Points = %d\n", pointset->NPoints());
    printf("  # Properties = 4\n");
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Program argument parsing
////////////////////////////////////////////////////////////////////////

static int 
ParseArgs(int argc, char **argv)
{
  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = 1;
      else if (!strcmp(*argv, "-debug")) print_debug = 1;
      else if (!strcmp(*argv, "-input_points")) { argv++; argc--; input_points_name = *argv; }
      else if (!strcmp(*argv, "-input_properties")) { argv++; argc--; input_properties_name = *argv; }
      else if (!strcmp(*argv, "-input_grid")) { argv++; argc--; input_grid_name = *argv; }
      else if (!strcmp(*argv, "-input_planes")) { argv++; argc--; input_planes_name = *argv; }
      else if (!strcmp(*argv, "-output_points")) { argv++; argc--; output_points_name = *argv; }
      else if (!strcmp(*argv, "-output_point_properties")) { argv++; argc--; output_point_properties_name = *argv; }
      else if (!strcmp(*argv, "-output_vertex_properties")) { argv++; argc--; output_vertex_properties_name = *argv; }
      else if (!strcmp(*argv, "-output_grid")) { argv++; argc--; output_grid_name = *argv; }
      else if (!strcmp(*argv, "-output_planes")) { argv++; argc--; output_planes_name = *argv; }
      else if (!strcmp(*argv, "-max_points")) { argv++; argc--; max_points = atoi(*argv); }
      else if (!strcmp(*argv, "-min_value")) { argv++; argc--; min_value = atof(*argv); }
      else if (!strcmp(*argv, "-min_weight")) { argv++; argc--; min_weight = atof(*argv); }
      else if (!strcmp(*argv, "-radius")) { argv++; argc--; radius = atof(*argv); }
      else if (!strcmp(*argv, "-in_plane_vector")) { 
        argv++; argc--; double vx = atof(*argv); 
        argv++; argc--; double vy = atof(*argv); 
        argv++; argc--; double vz = atof(*argv); 
        in_plane_vector.Reset(vx, vy, vz);
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0; 
      }
      argv++; argc--;
    }
    else {
      if (!input_mesh_name) input_mesh_name = *argv;
      else if (!output_planes_name) output_planes_name = *argv;
      else { fprintf(stderr, "Invalid program argument: %s", *argv); return 0; }
      argv++; argc--;
    }
  }

  // Check input filename
  if (!input_mesh_name || !output_planes_name) {
    fprintf(stderr, "Usage: msh2pln inputmeshfile outputplanesfile [options]\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Main program
////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(1);

  // Read mesh file
  R3Mesh *mesh = ReadMesh(input_mesh_name);
  if (!mesh) exit(-1);

  // Read/create points
  PointSet *pointset = NULL;
  if (input_points_name) pointset = ReadPoints(mesh, input_points_name);
  else pointset = CreatePoints(mesh, max_points);
  if (!pointset) exit(-1);

  // Read/create properties
  int property_status = 0;
  if (input_properties_name) property_status = ReadProperties(pointset, input_properties_name);
  else property_status = CreateProperties(pointset);
  if (!property_status) exit(-1);

  // Read/create grid
  R3Grid *grid = NULL;
  if (input_grid_name) grid = ReadGrid(input_grid_name);
  else grid = CreateGrid(pointset, radius);
  if (!grid) exit(-1);
  
  // Read/create planes
  PlaneSet *planeset = NULL;
  if (input_planes_name) planeset = ReadPlanes(mesh, input_planes_name);
  else planeset = CreatePlanes(pointset, grid, radius);
  if (!planeset) exit(-1);

  // Process planes
  if (!SortPlanes(planeset)) exit(-1);
  if (!UniquifyPlanes(planeset)) exit(-1);

  // Write points 
  if (output_points_name) {
    if (!WritePoints(pointset, output_points_name)) exit(-1);
  }
    
  // Write point properties 
  if (output_point_properties_name) {
    if (!WritePointProperties(pointset, output_point_properties_name)) exit(-1);
  }
    
  // Write properties 
  if (output_vertex_properties_name) {
    if (!WriteVertexProperties(pointset, output_vertex_properties_name)) exit(-1);
  }
    
  // Write grid 
  if (output_grid_name) {
    if (!WriteGrid(grid, output_grid_name)) exit(-1);
  }

  // Write planes 
  if (output_planes_name) {
    if (!WritePlanes(planeset, output_planes_name)) exit(-1);
  }
 
  // Return success 
  return 0;
}


