// Source file for the qslim program



// Include files 

#include "R3Shapes/R3Shapes.h"



// Program arguments

static char *input_mesh_name = NULL;
static char *output_mesh_name = NULL;
static RNScalar boundary_weight = 0.001;
static RNScalar max_error = FLT_MAX;
static int max_vertices = INT_MAX;
static int max_faces = INT_MAX;
static int keep_vertex_positions = 0;
static int print_verbose = 0;
static int print_debug = 0;



// Application data types

struct VertexData {
  R3MeshVertex *vertex;
  R4Matrix quadric;
  int original_id;
};

struct EdgeData {
  R3MeshEdge *edge;
  R4Matrix quadric;
  RNScalar quadric_error;
  RNScalar error;
  R3Point position;
  struct EdgeData **heapentry;
};



////////////////////////////////////////////////////////////////////////
// SIMPLIFICATION STUFF
////////////////////////////////////////////////////////////////////////

static int 
ComputeVertexQuadric(R3Mesh *mesh, R3MeshVertex *vertex)
{
  // Allocate vertex data
  VertexData *vdata = (VertexData *) mesh->VertexData(vertex);
  if (!vdata) {
    vdata = new VertexData();
    vdata->vertex = vertex;
    vdata->quadric = R4null_matrix;
    vdata->original_id = mesh->VertexID(vertex);
    mesh->SetVertexData(vertex, vdata);
  }

  // Sum quadric for all adjacent faces
  R3mesh_mark++;
  for (int j = 0; j < mesh->VertexValence(vertex); j++) {
    R3MeshEdge *edge = mesh->EdgeOnVertex(vertex, j);
    for (int k = 0; k < 2; k++) {
      R3MeshFace *face = mesh->FaceOnEdge(edge, k);
      if (face) {
        // Check if face has already been considered
        if (mesh->FaceMark(face) == R3mesh_mark) continue;
        mesh->SetFaceMark(face, R3mesh_mark);

        // Compute quadric for face
        const R3Plane& p = mesh->FacePlane(face);
        R4Matrix quadric(p[0]*p[0], p[0]*p[1], p[0]*p[2], p[0]*p[3],
                         p[1]*p[0], p[1]*p[1], p[1]*p[2], p[1]*p[3],
                         p[2]*p[0], p[2]*p[1], p[2]*p[2], p[2]*p[3],
                         p[3]*p[0], p[3]*p[1], p[3]*p[2], p[3]*p[3]);

        // Add to quadric for vertex
        vdata->quadric += mesh->FaceArea(face) * quadric;
      }
    }
  }

  // Add term for boundaries
  if (boundary_weight > 0) {
    for (int j = 0; j < mesh->VertexValence(vertex); j++) {
      R3MeshEdge *edge = mesh->EdgeOnVertex(vertex, j);
      for (int k = 0; k < 2; k++) {
        R3MeshFace *face = mesh->FaceOnEdge(edge, k);
        if (!face) {
          // Edge is on boundary
          face = mesh->FaceOnEdge(edge, 1-k);
          if (face) {
            // Add boundary
            R3MeshVertex *cw_vertex = mesh->VertexOnEdge(edge, face, RN_CW);
            R3MeshVertex *ccw_vertex = mesh->VertexOnEdge(edge, face, RN_CCW);
            const R3Point& cw_point = mesh->VertexPosition(cw_vertex);
            const R3Point& ccw_point = mesh->VertexPosition(ccw_vertex);
            R3Vector vector = ccw_point - cw_point;
            RNScalar length = vector.Length();
            if (length > RN_EPSILON) {
              vector /= length;
              vector.Cross(mesh->FaceNormal(face));
              R3Plane p(ccw_point, vector);
              R4Matrix quadric(p[0]*p[0], p[0]*p[1], p[0]*p[2], p[0]*p[3],
                               p[1]*p[0], p[1]*p[1], p[1]*p[2], p[1]*p[3],
                               p[2]*p[0], p[2]*p[1], p[2]*p[2], p[2]*p[3],
                               p[3]*p[0], p[3]*p[1], p[3]*p[2], p[3]*p[3]);

              // Add to quadric for vertex
              vdata->quadric += boundary_weight * quadric;
            }
          }
        }
      }
    }
  }

  // Return success
  return 1;
}



static int 
ComputeEdgeQuadric(R3Mesh *mesh, R3MeshEdge *edge, RNBoolean update_quadric_error)
{
  // Get vertices
  R3MeshVertex *vertex0 = mesh->VertexOnEdge(edge, 0);
  R3MeshVertex *vertex1 = mesh->VertexOnEdge(edge, 1);
  VertexData *vdata0 = (VertexData *) mesh->VertexData(vertex0);
  VertexData *vdata1 = (VertexData *) mesh->VertexData(vertex1);
  assert(vdata0 && vdata1);

  // Get convenient variables
  R3MeshFace *face0 = mesh->FaceOnEdge(edge, 0);
  R3MeshFace *face1 = mesh->FaceOnEdge(edge, 1);
  R3MeshEdge *edge00 = (face0) ? mesh->EdgeAcrossVertex(vertex0, edge, face0) : NULL;
  R3MeshEdge *edge10 = (face1) ? mesh->EdgeAcrossVertex(vertex0, edge, face1) : NULL;
  R3MeshEdge *edge01 = (face0) ? mesh->EdgeAcrossVertex(vertex1, edge, face0) : NULL;
  R3MeshEdge *edge11 = (face1) ? mesh->EdgeAcrossVertex(vertex1, edge, face1) : NULL;

  // Allocate edge data
  EdgeData *edata = (EdgeData *) mesh->EdgeData(edge);
  if (!edata) {
    edata = new EdgeData();
    edata->edge = edge;
    edata->quadric = R4null_matrix;
    edata->quadric_error = 0;
    edata->error = 0;
    edata->position = R3zero_point;
    edata->heapentry = NULL;
    mesh->SetEdgeData(edge, edata);
  }

  // Check if need to update quadric error
  if (update_quadric_error) {
    // Compute quadric
    edata->quadric = vdata0->quadric + vdata1->quadric;

    // Compute position of future new vertex
    R4Matrix inverse = edata->quadric;
    inverse[3][0] = 0;
    inverse[3][1] = 0;
    inverse[3][2] = 0;
    inverse[3][3] = 1;
    if (keep_vertex_positions || RNIsZero(inverse.Determinant())) {
      // Choose best among edge endpoints
      R3Vector v0 = mesh->VertexPosition(vertex0).Vector();
      RNScalar e0 = v0.Dot(edata->quadric * v0);
      R3Vector v1 = mesh->VertexPosition(vertex1).Vector();
      RNScalar e1 = v1.Dot(edata->quadric * v1);
      edata->position = (e0 < e1) ? v0.Point() : v1.Point();
    }
    else {
      // Choose position that minimizes QEM
      inverse.Invert();
      edata->position.Reset(inverse[0][3], inverse[1][3], inverse[2][3]);
    }

    // Compute error
    RNScalar vQv = 0;
    R3Point& p = edata->position;
    R4Matrix& m = edata->quadric;
    vQv += p[0]*(m[0][0]*p[0] + m[0][1]*p[1] + m[0][2]*p[2] + m[0][3]);
    vQv += p[1]*(m[1][0]*p[0] + m[1][1]*p[1] + m[1][2]*p[2] + m[1][3]);
    vQv += p[2]*(m[2][0]*p[0] + m[2][1]*p[1] + m[2][2]*p[2] + m[2][3]);
    vQv +=      (m[3][0]*p[0] + m[3][1]*p[1] + m[3][2]*p[2] + m[3][3]);
    edata->quadric_error = vQv;
  }

  // Update error
  edata->error = edata->quadric_error;

  // Check for back-to-back triangles
  if (edge00 && edge10) {
    R3MeshVertex *v00 = mesh->VertexAcrossEdge(edge00, vertex0);
    R3MeshVertex *v10 = mesh->VertexAcrossEdge(edge10, vertex0);
    if (v00 == v10) edata->error += RN_INFINITY;
  }
  else if (edge11 && edge01) {
    R3MeshVertex *v11 = mesh->VertexAcrossEdge(edge11, vertex1);
    R3MeshVertex *v01 = mesh->VertexAcrossEdge(edge01, vertex1);
    if (v11 == v01) edata->error += RN_INFINITY;
  }

  // Check for fin
  for (int i = 0; i < mesh->VertexValence(vertex0); i++) {
    R3MeshEdge *edge0 = mesh->EdgeOnVertex(vertex0, i);
    if ((edge0 == edge00) || (edge0 == edge10)) continue;
    R3MeshVertex *v0v = mesh->VertexAcrossEdge(edge0, vertex0);
    for (int j = 0; j < mesh->VertexValence(vertex1); j++) {
      R3MeshEdge *edge1 = mesh->EdgeOnVertex(vertex1, j);
      if ((edge1 == edge01) || (edge1 == edge11)) continue;
      R3MeshVertex *v1v = mesh->VertexAcrossEdge(edge1, vertex1);
      if (v0v == v1v) {
        edata->error += RN_INFINITY;
      }
    }
  }

#if 0
  // Check for fold over on vertex0 side
  for (int i = 0; i < mesh->VertexValence(vertex0); i++) {
    R3MeshEdge *e = mesh->EdgeOnVertex(vertex0, i);
    if ((e == edge00) || (e == edge10)) continue;
    for (int j = 0; j < 2; j++) {
      R3MeshFace *f = mesh->FaceOnEdge(e, j);
      if (f) {
        const R3Point& p0 = mesh->VertexPosition(vertex0);
        const R3Point& p1 = mesh->VertexPosition(mesh->VertexOnFace(f, vertex0, RN_CCW));
        const R3Point& p2 = mesh->VertexPosition(mesh->VertexOnFace(f, vertex0, RN_CW));
        R3Vector edge_vector = p2 - p1;
        edge_vector.Normalize();
        R3Vector edge_normal = mesh->FaceNormal(f) % edge_vector;
        edge_normal.Normalize();
        R3Plane edge_plane(p2, edge_normal);
        RNScalar d_before = R3SignedDistance(edge_plane, p0);
        RNScalar d_after = R3SignedDistance(edge_plane, edata->position);
        if ((d_before * d_after) < 0) {
          static const RNScalar penalty = sqrt(RN_INFINITY);
          edata->error += penalty;
        }
      }
    }
  }

  // Check for fold over on vertex1 side
  for (int i = 0; i < mesh->VertexValence(vertex1); i++) {
    R3MeshEdge *e = mesh->EdgeOnVertex(vertex1, i);
    if ((e == edge00) || (e == edge10)) continue;
    for (int j = 0; j < 2; j++) {
      R3MeshFace *f = mesh->FaceOnEdge(e, j);
      if (f) {
        const R3Point& p0 = mesh->VertexPosition(vertex1);
        const R3Point& p1 = mesh->VertexPosition(mesh->VertexOnFace(f, vertex1, RN_CCW));
        const R3Point& p2 = mesh->VertexPosition(mesh->VertexOnFace(f, vertex1, RN_CW));
        R3Vector edge_vector = p2 - p1;
        edge_vector.Normalize();
        R3Vector edge_normal = mesh->FaceNormal(f) % edge_vector;
        edge_normal.Normalize();
        R3Plane edge_plane(p2, edge_normal);
        RNScalar d_before = R3SignedDistance(edge_plane, p0);
        RNScalar d_after = R3SignedDistance(edge_plane, edata->position);
        if ((d_before * d_after) < 0) {
          static const RNScalar penalty = sqrt(RN_INFINITY);
          edata->error += penalty;
        }
      }
    }
  }
#endif

  // Return success
  return 1;
}



static int 
SimplifyMesh(R3Mesh *mesh, RNScalar max_error, int max_vertices, int max_faces)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  RNScalar last_error = 0;

  // Compute quadric for every vertex
  for (int i = 0; i < mesh->NVertices(); i++) {
    R3MeshVertex *vertex = mesh->Vertex(i);
    if (!ComputeVertexQuadric(mesh, vertex)) return 0;
  }

  // Compute quadric for every edge
  for (int i = 0; i < mesh->NEdges(); i++) {
    R3MeshEdge *edge = mesh->Edge(i);
    if (!ComputeEdgeQuadric(mesh, edge, TRUE)) return 0;
  }

  // Build priority queue
  EdgeData tmp;
  RNHeap<EdgeData *> heap(&tmp, &tmp.error, &tmp.heapentry);
  for (int i = 0; i < mesh->NEdges(); i++) {
    R3MeshEdge *edge = mesh->Edge(i);
    EdgeData *edata = (EdgeData *) mesh->EdgeData(edge);
    heap.Push(edata);
  }

  // Collapse edges
  while ((mesh->NVertices() > max_vertices) || (mesh->NFaces() > max_faces)) {
    EdgeData *edata = heap.Pop();
    if (!edata) break;

    // Get edge
    R3MeshEdge *edge = edata->edge;
    assert(edge);

#ifndef NDEBUG
    // Check edge error
    RNScalar error = edata->error;
    ComputeEdgeQuadric(mesh, edge, TRUE);
    if (edata->error != error) {
      printf("Mismatched error: %g %g\n", error, edata->error);
      heap.Push(edata);
      continue;
    }

    // Check if edge is in mesh
    if (!mesh->IsEdgeOnMesh(edge)) {
      printf("Trying to collapse edge that is not on mesh: %d\n", mesh->EdgeID(edge));
      continue;
    }
#endif

    // Check if we are done
    if (edata->error >= RN_INFINITY) break;
    if (edata->error > max_error) break;
    last_error = edata->error;

    // Track vertex IDs as collapse
    if (1) {
      // Swap original vertex IDs if v1 is closer
      R3MeshVertex *v0 = mesh->VertexOnEdge(edge, 0);
      R3MeshVertex *v1 = mesh->VertexOnEdge(edge, 1);
      VertexData *vdata0 = (VertexData *) mesh->VertexData(v0);
      VertexData *vdata1 = (VertexData *) mesh->VertexData(v1);
      const R3Point& p0 = mesh->VertexPosition(v0);
      const R3Point& p1 = mesh->VertexPosition(v1);
      RNLength d0 = R3SquaredDistance(p0, edata->position);
      RNLength d1 = R3SquaredDistance(p1, edata->position);
      if (d1 < d0) {
        int swap = vdata0->original_id ;
        vdata0->original_id = vdata1->original_id;
        vdata1->original_id = swap;
      }
    }

    // Print a debug message 
    if (print_debug) {
      R3MeshVertex *v1 = mesh->VertexOnEdge(edge, 1);
      VertexData *vdata1 = (VertexData *) mesh->VertexData(v1);
      int original_id1 = vdata1->original_id;
      RNArea area = 0;
      R3MeshFace *f0 = mesh->FaceOnEdge(edge, 0);
      if (f0) area += mesh->FaceArea(f0);
      R3MeshFace *f1 = mesh->FaceOnEdge(edge, 1);
      if (f1) area += mesh->FaceArea(f1);
      printf("  %d = vertexid    %d = nfaces    %d = nverts    %g = error    %g = area\n", 
        original_id1, mesh->NFaces(), mesh->NVertices(), edata->error, area);
    }

    // Remove edges to be deleted from queue (e01, e11)
    R3MeshVertex *v1 = mesh->VertexOnEdge(edge, 1);
    R3MeshFace *f0 = mesh->FaceOnEdge(edge, 0);  
    R3MeshFace *f1 = mesh->FaceOnEdge(edge, 1);  
    R3MeshEdge *e01 = (f0) ? mesh->EdgeAcrossVertex(v1, edge, f0) : NULL;
    R3MeshEdge *e11 = (f1) ? mesh->EdgeAcrossVertex(v1, edge, f1) : NULL;
    EdgeData *edata01 = (e01) ? (EdgeData *) mesh->EdgeData(e01) : NULL;
    EdgeData *edata11 = (e11) ? (EdgeData *) mesh->EdgeData(e11) : NULL;
    if (edata01) heap.Remove(edata01);
    if (edata11) heap.Remove(edata11);

    // Collapse edge
    R3MeshVertex *vertex = mesh->CollapseEdge(edge, edata->position);
    if (!vertex) continue;

    // Update vertex quadric
    VertexData *vdata = (VertexData *) mesh->VertexData(vertex);
    assert(vdata && (vdata->vertex == vertex));
    vdata->quadric = edata->quadric;

    // Update edge quadrics in a two-ring neighborhood
    R3mesh_mark++;
    for (int j = 0; j < mesh->VertexValence(vertex); j++) {
      R3MeshEdge *e = mesh->EdgeOnVertex(vertex, j);
      if (mesh->EdgeMark(e) != R3mesh_mark) {
        mesh->SetEdgeMark(e, R3mesh_mark);
        ComputeEdgeQuadric(mesh, e, TRUE);
        EdgeData *ed = (EdgeData *) mesh->EdgeData(e);
        assert(ed);
        heap.Update(ed);
      }
      R3MeshVertex *ev = mesh->VertexAcrossEdge(e, vertex);
      for (int k = 0; k < mesh->VertexValence(ev); k++) {
        R3MeshEdge *eve = mesh->EdgeOnVertex(ev, k);
        if (mesh->EdgeMark(eve) != R3mesh_mark) {
          mesh->SetEdgeMark(eve, R3mesh_mark);
          ComputeEdgeQuadric(mesh, eve, FALSE);
          EdgeData *eved = (EdgeData *) mesh->EdgeData(eve);
          assert(eved);
          heap.Update(eved);
        }
      }
    }
  }

  // Print final debug message with point ids
  if (print_debug) {
    for (int i = 0; i < mesh->NVertices(); i++) {
      R3MeshVertex *v = mesh->Vertex(i);
      VertexData *vdata = (VertexData *) mesh->VertexData(v);
      const R3Point& p = mesh->VertexPosition(v);
      RNArea area = mesh->VertexArea(v);
      R4Matrix& m = vdata->quadric;
      RNScalar vQv = 0;
      vQv += p[0]*(m[0][0]*p[0] + m[0][1]*p[1] + m[0][2]*p[2] + m[0][3]);
      vQv += p[1]*(m[1][0]*p[0] + m[1][1]*p[1] + m[1][2]*p[2] + m[1][3]);
      vQv += p[2]*(m[2][0]*p[0] + m[2][1]*p[1] + m[2][2]*p[2] + m[2][3]);
      vQv +=      (m[3][0]*p[0] + m[3][1]*p[1] + m[3][2]*p[2] + m[3][3]);
      RNScalar error = vQv;
      printf("  %d = vertexid    %d = nfaces    %d = nverts    %g = error    %g = area\n",
             vdata->original_id, mesh->NFaces(), mesh->NVertices(), error, area);
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Simplified mesh ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Faces = %d\n", mesh->NFaces());
    printf("  # Edges = %d\n", mesh->NEdges());
    printf("  # Vertices = %d\n", mesh->NVertices());
    printf("  Max Faces = %d\n", (max_faces != INT_MAX) ? max_faces : 0);
    printf("  Max Vertices = %d\n", (max_vertices != INT_MAX) ? max_vertices : 0);
    printf("  Max Error = %g\n", (max_error != FLT_MAX) ? max_error : 0);
    printf("  Error = %g\n", last_error);
    fflush(stdout);
  }

  // Check if mesh is valid
  assert(mesh->IsValid());

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// I/O STUFF
////////////////////////////////////////////////////////////////////////

static R3Mesh *
ReadMesh(char *mesh_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate mesh
  R3Mesh *mesh = new R3Mesh();
  assert(mesh);

  // Read mesh from file
  if (!mesh->ReadFile(mesh_name)) {
    delete mesh;
    return NULL;
  }

  // Check if mesh is valid
  assert(mesh->IsValid());

  // Print statistics
  if (print_verbose) {
    printf("Read mesh ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Faces = %d\n", mesh->NFaces());
    printf("  # Edges = %d\n", mesh->NEdges());
    printf("  # Vertices = %d\n", mesh->NVertices());
    fflush(stdout);
  }

  // Return success
  return mesh;
}



static int
WriteMesh(R3Mesh *mesh, char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write mesh to file
  if (!mesh->WriteFile(filename)) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Wrote mesh to %s ...\n", filename);
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Faces = %d\n", mesh->NFaces());
    printf("  # Edges = %d\n", mesh->NEdges());
    printf("  # Vertices = %d\n", mesh->NVertices());
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// ARGUMENT PARSING STUFF
////////////////////////////////////////////////////////////////////////

static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc == 1) {
    printf("Usage: qslim inputname outputname [options]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = 1; 
      else if (!strcmp(*argv, "-debug")) print_debug = 1; 
      else if (!strcmp(*argv, "-keep_vertex_positions")) keep_vertex_positions = 1; 
      else if (!strcmp(*argv, "-max_error")) { argv++; argc--; max_error = atof(*argv); }
      else if (!strcmp(*argv, "-max_faces")) { argv++; argc--; max_faces = atoi(*argv); }
      else if (!strcmp(*argv, "-max_vertices")) { argv++; argc--; max_vertices = atoi(*argv); }
      else if (!strcmp(*argv, "-boundary_weight")) { argv++; argc--; boundary_weight = 0.001 * atof(*argv); }
      else { fprintf(stderr, "Invalid program argument: %s\n", *argv); exit(1); }
      argv++; argc--;
    }
    else {
      if (!input_mesh_name) input_mesh_name = *argv;
      else if (!output_mesh_name) output_mesh_name = *argv;
      else { fprintf(stderr, "Invalid program argument: %s", *argv); exit(1); }
      argv++; argc--;
    }
  }

  // Check input filename
  if (!input_mesh_name || !output_mesh_name) {
    fprintf(stderr, "Usage: qslim inputmesh outputmesh [options]\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



////////////////////////////////////////////////////////////////////////
// MAIN
////////////////////////////////////////////////////////////////////////

int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read mesh
  R3Mesh *mesh = ReadMesh(input_mesh_name);
  if (!mesh) exit(-1);

  // Simplify mesh
  if (!SimplifyMesh(mesh, max_error, max_vertices, max_faces)) exit(-1);

  // Write mesh
  if (!WriteMesh(mesh, output_mesh_name)) exit(-1);

  // Return success 
  return 0;
}

















