// Source file for the qslim program

// #define ALEX_LIGHTS



// Include files 

#include "R3Graphics/R3Graphics.h"
#include "fglut/fglut.h"
#include "GL/glui.h"



// Program arguments

static char *input_mesh_name = NULL;
static char *output_mesh_name = NULL;
static char *input_correspondences_name = NULL;
static char *output_correspondences_name = NULL;
static char *output_image_name = NULL;
static R3Point initial_camera_origin = R3Point(0.0, 0.0, 0.0);
static R3Vector initial_camera_towards = R3Vector(0.0, 0.0, -1.0);
static R3Vector initial_camera_up = R3Vector(0.0, 1.0, 0.0);
static RNBoolean initial_camera = FALSE;
static RNArray<R3Plane *> planes;
static RNScalar boundary_weight = 0.001;
static int nfaces = INT_MAX;
static int interactive = 0;
static int print_verbose = 0;



// GLUI commands

enum {
  NULL_COMMAND,
  QUIT_COMMAND
};



// GLUI variables 

static GLUI *glui = NULL;
static int GLUIwindow = 0;
static int GLUIwindow_height = 480;
static int GLUIwindow_width = 640;
static int GLUImouse[2] = { 0, 0 };
static int GLUIbutton[3] = { 0, 0, 0 };
static int GLUImodifiers = 0;




// Display variables

static int rocking = 0;
static int show_faces = 1;
static int show_edges = 0;
static int show_vertices = 0;
static int show_planes = 0;
static int show_backfacing = 0;
static int show_gouraud = 0;
static int show_axes = 0;
static int show_names = 0;
static int show_reflected_faces = 0;
static int show_face_correspondences = 0;
static int show_edge_correspondences = 0;
static int show_vertex_correspondences = 0;
static int show_errors = 0;
static RNScalar show_errors_scale = 1;
static int save_image = 0;
static const int ncolors = 18;
static GLfloat colors[ncolors][4] = { 
  {1,0,0,1}, {0,1,0,1}, {0,0,1,1}, {1,1,0,1}, 
  {1,0,1,1}, {0,1,1,1}, {1,0.5,0,1}, {1,0,0.5,1}, 
  {0.5,1,0,1}, {0.5,0,1,1}, {0,0.5,1,1}, {0,1,0.5,1}, 
  {1,0.7,0.3,1}, {1,0.3,0.7,1}, {0.3,1,0.7,1}, {0.3,0.7,1,1}, 
  {0.7,1,0.3,1}, {0.7,0.3,1,1} };


// Application data types

struct VertexCorrespondence {
  RNArray<R3MeshVertex *> vertices;
};

struct EdgeCorrespondence {
  RNArray<R3MeshEdge *> edges;
  RNScalar error;
  struct EdgeCorrespondence **heapentry;
  RNMark mark;
};

struct FaceCorrespondence {
  RNArray<R3MeshFace *> faces;
};

struct VertexData {
  R3MeshVertex *vertex;
  R4Matrix quadric;
  VertexCorrespondence *correspondence;
};

struct EdgeData {
  R3MeshEdge *edge;
  R4Matrix quadric;
  RNScalar quadric_error;
  RNScalar error;
  R3Point position;
  int id;
  struct EdgeData **heapentry;
  EdgeCorrespondence *correspondence;
};

struct FaceData {
  R3MeshFace *face;
  FaceCorrespondence *correspondence;
};



// Application variables

static R3Viewer *viewer = NULL;
static R3Mesh *mesh = NULL;
static RNArray<VertexCorrespondence *> *vertex_correspondences = NULL;
static RNArray<EdgeCorrespondence *> *edge_correspondences = NULL;
static RNArray<FaceCorrespondence *> *face_correspondences = NULL;
static RNScalar scale = 1;



////////////////////////////////////////////////////////////////////////
// SIMPLIFICATION STUFF
////////////////////////////////////////////////////////////////////////

static int 
ComputeVertexQuadric(R3Mesh *mesh, R3MeshVertex *vertex)
{
  // Allocate vertex data
  VertexData *vdata = (VertexData *) mesh->VertexData(vertex);
  if (!vdata) {
    vdata = new VertexData();
    vdata->vertex = vertex;
    vdata->quadric = R4null_matrix;
    vdata->correspondence = NULL;
    mesh->SetVertexData(vertex, vdata);
  }

  // Sum quadric for all adjacent faces
  R3mesh_mark++;
  for (int j = 0; j < mesh->VertexValence(vertex); j++) {
    R3MeshEdge *edge = mesh->EdgeOnVertex(vertex, j);
    for (int k = 0; k < 2; k++) {
      R3MeshFace *face = mesh->FaceOnEdge(edge, k);
      if (face) {
        // Check if face has already been considered
        if (mesh->FaceMark(face) == R3mesh_mark) continue;
        mesh->SetFaceMark(face, R3mesh_mark);

        // Compute quadric for face
        const R3Plane& p = mesh->FacePlane(face);
        R4Matrix quadric(p[0]*p[0], p[0]*p[1], p[0]*p[2], p[0]*p[3],
                         p[1]*p[0], p[1]*p[1], p[1]*p[2], p[1]*p[3],
                         p[2]*p[0], p[2]*p[1], p[2]*p[2], p[2]*p[3],
                         p[3]*p[0], p[3]*p[1], p[3]*p[2], p[3]*p[3]);

        // Add to quadric for vertex
        vdata->quadric += mesh->FaceArea(face) * quadric;
      }
    }
  }

  // Add term for boundaries
  if (boundary_weight > 0) {
    for (int j = 0; j < mesh->VertexValence(vertex); j++) {
      R3MeshEdge *edge = mesh->EdgeOnVertex(vertex, j);
      for (int k = 0; k < 2; k++) {
        R3MeshFace *face = mesh->FaceOnEdge(edge, k);
        if (!face) {
          // Edge is on boundary
          face = mesh->FaceOnEdge(edge, 1-k);
          if (face) {
            // Add boundary
            R3MeshVertex *cw_vertex = mesh->VertexOnEdge(edge, face, RN_CW);
            R3MeshVertex *ccw_vertex = mesh->VertexOnEdge(edge, face, RN_CCW);
            const R3Point& cw_point = mesh->VertexPosition(cw_vertex);
            const R3Point& ccw_point = mesh->VertexPosition(ccw_vertex);
            R3Vector vector = ccw_point - cw_point;
            RNScalar length = vector.Length();
            if (length > RN_EPSILON) {
              vector /= length;
              vector.Cross(mesh->FaceNormal(face));
              R3Plane p(ccw_point, vector);
              R4Matrix quadric(p[0]*p[0], p[0]*p[1], p[0]*p[2], p[0]*p[3],
                               p[1]*p[0], p[1]*p[1], p[1]*p[2], p[1]*p[3],
                               p[2]*p[0], p[2]*p[1], p[2]*p[2], p[2]*p[3],
                               p[3]*p[0], p[3]*p[1], p[3]*p[2], p[3]*p[3]);

              // Add to quadric for vertex
              vdata->quadric += boundary_weight * scale * scale * quadric;
            }
          }
        }
      }
    }
  }

  // Return success
  return 1;
}



static int 
ComputeEdgeQuadric(R3Mesh *mesh, R3MeshEdge *edge, RNBoolean update_quadric_error)
{
  // Get vertices
  R3MeshVertex *vertex0 = mesh->VertexOnEdge(edge, 0);
  R3MeshVertex *vertex1 = mesh->VertexOnEdge(edge, 1);
  VertexData *vdata0 = (VertexData *) mesh->VertexData(vertex0);
  VertexData *vdata1 = (VertexData *) mesh->VertexData(vertex1);
  assert(vdata0 && vdata1);

  // Allocate edge data
  EdgeData *edata = (EdgeData *) mesh->EdgeData(edge);
  if (!edata) {
    edata = new EdgeData();
    edata->edge = edge;
    edata->quadric = R4null_matrix;
    edata->quadric_error = 0;
    edata->error = 0;
    edata->position = R3zero_point;
    edata->id = mesh->EdgeID(edge);
    edata->heapentry = NULL;
    edata->correspondence = NULL;
    mesh->SetEdgeData(edge, edata);
  }

  // Check if need to update quadric error
  if (update_quadric_error) {
    // Compute quadric
    edata->quadric = vdata0->quadric + vdata1->quadric;

    // Compute position of future new vertex
    R4Matrix inverse = edata->quadric;
    inverse[3][0] = 0;
    inverse[3][1] = 0;
    inverse[3][2] = 0;
    inverse[3][3] = 1;
    if (RNIsNotZero(inverse.Determinant())) {
      // Choose position that minimizes QEM
      inverse.Invert();
      edata->position.Reset(inverse[0][3], inverse[1][3], inverse[2][3]);
    }
    else {
      // Choose best among edge endpoints 
      R3Vector v0 = mesh->VertexPosition(vertex0).Vector();
      RNScalar e0 = v0.Dot(edata->quadric * v0);
      R3Vector v1 = mesh->VertexPosition(vertex1).Vector();
      RNScalar e1 = v1.Dot(edata->quadric * v1);
      edata->position = (e0 < e1) ? v0.Point() : v1.Point();
    }

    // Compute error
    RNScalar vQv = 0;
    R3Point& p = edata->position;
    R4Matrix& m = edata->quadric;
    vQv += p[0]*(m[0][0]*p[0] + m[0][1]*p[1] + m[0][2]*p[2] + m[0][3]);
    vQv += p[1]*(m[1][0]*p[0] + m[1][1]*p[1] + m[1][2]*p[2] + m[1][3]);
    vQv += p[2]*(m[2][0]*p[0] + m[2][1]*p[1] + m[2][2]*p[2] + m[2][3]);
    vQv +=      (m[3][0]*p[0] + m[3][1]*p[1] + m[3][2]*p[2] + m[3][3]);
    edata->quadric_error = vQv;
  }

  // Update error
  edata->error = edata->quadric_error;

  // Check for fin
  R3MeshFace *face0 = mesh->FaceOnEdge(edge, 0);
  R3MeshFace *face1 = mesh->FaceOnEdge(edge, 1);
  R3MeshEdge *edge00 = (face0) ? mesh->EdgeAcrossVertex(vertex0, edge, face0) : NULL;
  R3MeshEdge *edge10 = (face1) ? mesh->EdgeAcrossVertex(vertex0, edge, face1) : NULL;
  R3MeshEdge *edge01 = (face0) ? mesh->EdgeAcrossVertex(vertex1, edge, face0) : NULL;
  R3MeshEdge *edge11 = (face1) ? mesh->EdgeAcrossVertex(vertex1, edge, face1) : NULL;
  for (int i = 0; i < mesh->VertexValence(vertex0); i++) {
    R3MeshEdge *edge0 = mesh->EdgeOnVertex(vertex0, i);
    if ((edge0 == edge00) || (edge0 == edge10)) continue;
    R3MeshVertex *v0v = mesh->VertexAcrossEdge(edge0, vertex0);
    for (int j = 0; j < mesh->VertexValence(vertex1); j++) {
      R3MeshEdge *edge1 = mesh->EdgeOnVertex(vertex1, j);
      if ((edge1 == edge01) || (edge1 == edge11)) continue;
      R3MeshVertex *v1v = mesh->VertexAcrossEdge(edge1, vertex1);
      if (v0v == v1v) {
        edata->error += RN_INFINITY;
      }
    }
  }

#if 0
  // Check for fold over on vertex0 side
  for (int i = 0; i < mesh->VertexValence(vertex0); i++) {
    R3MeshEdge *e = mesh->EdgeOnVertex(vertex0, i);
    if ((e == edge00) || (e == edge10)) continue;
    for (int j = 0; j < 2; j++) {
      R3MeshFace *f = mesh->FaceOnEdge(e, j);
      if (f) {
        const R3Point& p0 = mesh->VertexPosition(vertex0);
        const R3Point& p1 = mesh->VertexPosition(mesh->VertexOnFace(f, vertex0, RN_CCW));
        const R3Point& p2 = mesh->VertexPosition(mesh->VertexOnFace(f, vertex0, RN_CW));
        R3Vector edge_vector = p2 - p1;
        edge_vector.Normalize();
        R3Vector edge_normal = mesh->FaceNormal(f) % edge_vector;
        edge_normal.Normalize();
        R3Plane edge_plane(p2, edge_normal);
        RNScalar d_before = R3SignedDistance(edge_plane, p0);
        RNScalar d_after = R3SignedDistance(edge_plane, edata->position);
        if ((d_before * d_after) < 0) {
          static const RNScalar penalty = sqrt(RN_INFINITY);
          edata->error += penalty;
        }
      }
    }
  }

  // Check for fold over on vertex1 side
  for (int i = 0; i < mesh->VertexValence(vertex1); i++) {
    R3MeshEdge *e = mesh->EdgeOnVertex(vertex1, i);
    if ((e == edge00) || (e == edge10)) continue;
    for (int j = 0; j < 2; j++) {
      R3MeshFace *f = mesh->FaceOnEdge(e, j);
      if (f) {
        const R3Point& p0 = mesh->VertexPosition(vertex1);
        const R3Point& p1 = mesh->VertexPosition(mesh->VertexOnFace(f, vertex1, RN_CCW));
        const R3Point& p2 = mesh->VertexPosition(mesh->VertexOnFace(f, vertex1, RN_CW));
        R3Vector edge_vector = p2 - p1;
        edge_vector.Normalize();
        R3Vector edge_normal = mesh->FaceNormal(f) % edge_vector;
        edge_normal.Normalize();
        R3Plane edge_plane(p2, edge_normal);
        RNScalar d_before = R3SignedDistance(edge_plane, p0);
        RNScalar d_after = R3SignedDistance(edge_plane, edata->position);
        if ((d_before * d_after) < 0) {
          static const RNScalar penalty = sqrt(RN_INFINITY);
          edata->error += penalty;
        }
      }
    }
  }
#endif

  // Return success
  return 1;
}



static int 
ComputeEdgeCorrespondenceError(R3Mesh *mesh, EdgeCorrespondence *correspondence)
{
  // Initialize error
  correspondence->error = 0;

  // Check if correspondence has one edge on symmetry boundary
  if (correspondence->edges.NEntries() == 1) {
    R3MeshEdge *edge = correspondence->edges.Kth(0);
    EdgeData *edata = (EdgeData *) mesh->EdgeData(edge);
    assert(edata);
    if (planes.IsEmpty()) edata->position.Project(mesh->EdgeLine(edge));
    else edata->position.Project(*(planes[0]));
    ComputeEdgeQuadric(mesh, edge, FALSE);
  }

  // Check if correspondence shares a vertex on symmetry boundary
  if (correspondence->edges.NEntries() == 2) {
    R3MeshEdge *e0 = correspondence->edges.Kth(0);
    R3MeshEdge *e1 = correspondence->edges.Kth(1);
    R3MeshVertex *v = mesh->VertexBetweenEdges(e0, e1);
    if (v) {
      EdgeData *e0data = (EdgeData *) mesh->EdgeData(e0);
      assert(e0data);
      e0data->position = mesh->VertexPosition(v);
      if (!planes.IsEmpty()) e0data->position.Project(*(planes[0]));
      ComputeEdgeQuadric(mesh, e0, FALSE);
      EdgeData *e1data = (EdgeData *) mesh->EdgeData(e1);
      assert(e1data);
      e1data->position = mesh->VertexPosition(v);
      if (!planes.IsEmpty()) e1data->position.Project(*(planes[0]));
      ComputeEdgeQuadric(mesh, e1, FALSE);
    }
  }

  // Compute maximum of edge errors
  for (int i = 0; i < correspondence->edges.NEntries(); i++) {
    R3MeshEdge *edge = correspondence->edges.Kth(i);
    EdgeData *edata = (EdgeData *) mesh->EdgeData(edge);
    assert(edata);
    if (edata->error > correspondence->error) {
      correspondence->error = edata->error;
    }
  }

  // Check for fin from combined collapse
  if (correspondence->error < RN_INFINITY) {
    // Check if corresponding edges share a vertex and are both adjacent to another vertex
    if (correspondence->edges.NEntries() > 1) {
      for (int i = 0; i < correspondence->edges.NEntries(); i++) {
        R3MeshEdge *edge1 = correspondence->edges.Kth(i);
        for (int j = i+1; j < correspondence->edges.NEntries(); j++) {
          R3MeshEdge *edge2 = correspondence->edges.Kth(j);
          assert(edge1 != edge2);
          R3MeshVertex *vertex = mesh->VertexBetweenEdges(edge1, edge2);
          if (vertex) {
            R3MeshVertex *vertex1 = mesh->VertexAcrossEdge(edge1, vertex);
            R3MeshVertex *vertex2 = mesh->VertexAcrossEdge(edge2, vertex);
            if (vertex1 != vertex2) {
              for (int r = 0; r < mesh->VertexValence(vertex1); r++) {
                R3MeshEdge *e1 = mesh->EdgeOnVertex(vertex1, r);
                R3MeshVertex *v1 = mesh->VertexAcrossEdge(e1, vertex1);
                if (v1 == vertex) continue;
                if (v1 == vertex2) { correspondence->error = RN_INFINITY; return 1; }
                for (int s = 0; s < mesh->VertexValence(vertex2); s++) {
                  R3MeshEdge *e2 = mesh->EdgeOnVertex(vertex2, s);
                  R3MeshVertex *v2 = mesh->VertexAcrossEdge(e2, vertex2);
                  if (v2 == vertex) continue;
                  if (v2 == vertex1) { correspondence->error = RN_INFINITY; return 1; }
                  if (v1 == v2) {correspondence->error = RN_INFINITY; return 1; }
                }
              }
            }
          }
        }
      }
    }
  }

  // Return success
  return 1;
}



static int 
SimplifyMesh(R3Mesh *mesh, int nfaces)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Compute quadric for every vertex
  for (int i = 0; i < mesh->NVertices(); i++) {
    R3MeshVertex *vertex = mesh->Vertex(i);
    if (!ComputeVertexQuadric(mesh, vertex)) return 0;
  }

  // Compute quadric for every edge
  for (int i = 0; i < mesh->NEdges(); i++) {
    R3MeshEdge *edge = mesh->Edge(i);
    if (!ComputeEdgeQuadric(mesh, edge, TRUE)) return 0;
  }

  // Build priority queue
  EdgeData tmp;
  RNHeap<EdgeData *> heap(&tmp, &tmp.error, &tmp.heapentry);
  for (int i = 0; i < mesh->NEdges(); i++) {
    R3MeshEdge *edge = mesh->Edge(i);
    EdgeData *edata = (EdgeData *) mesh->EdgeData(edge);
    heap.Push(edata);
  }

  // Collapse edges
  while (mesh->NFaces() > nfaces) {
    EdgeData *edata = heap.Pop();
    if (!edata) break;

    // Get edge
    R3MeshEdge *edge = edata->edge;
    assert(edge);

#ifndef NDEBUG
    // Check edge error
    RNScalar error = edata->error;
    ComputeEdgeQuadric(mesh, edge, TRUE);
    if (edata->error != error) {
      printf("Mismatched error: %g %g\n", error, edata->error);
      heap.Push(edata);
      continue;
    }

    // Check if edge is in mesh
    if (!mesh->IsEdgeOnMesh(edge)) {
      printf("Trying to collapse edge that is not on mesh: %d\n", mesh->EdgeID(edge));
      continue;
    }
#endif

    // Check if we are done
    if (edata->error >= RN_INFINITY) break;

    // Print a message every once in a while
    if (print_verbose) {
      static int next_nfaces = 0;
      if (next_nfaces == 0) {
        next_nfaces = 1000 * (mesh->NFaces() / 1000);
      }
      else if (mesh->NFaces() < next_nfaces) { 
        RNArea area = 0;
        R3MeshFace *f0 = mesh->FaceOnEdge(edge, 0);
        if (f0) area += mesh->FaceArea(f0);
        R3MeshFace *f1 = mesh->FaceOnEdge(edge, 1);
        if (f1) area += mesh->FaceArea(f1);
        printf("%d faces, %g error, %g area\n", mesh->NFaces(), edata->error, area);
        next_nfaces -= 1000;
      }
    }

    // Remove edges to be deleted from queue (e01, e11)
    R3MeshVertex *v1 = mesh->VertexOnEdge(edge, 1);
    R3MeshFace *f0 = mesh->FaceOnEdge(edge, 0);  
    R3MeshFace *f1 = mesh->FaceOnEdge(edge, 1);  
    R3MeshEdge *e01 = (f0) ? mesh->EdgeAcrossVertex(v1, edge, f0) : NULL;
    R3MeshEdge *e11 = (f1) ? mesh->EdgeAcrossVertex(v1, edge, f1) : NULL;
    EdgeData *edata01 = (e01) ? (EdgeData *) mesh->EdgeData(e01) : NULL;
    EdgeData *edata11 = (e11) ? (EdgeData *) mesh->EdgeData(e11) : NULL;
    if (edata01) heap.Remove(edata01);
    if (edata11) heap.Remove(edata11);

    // Collapse edge
    R3MeshVertex *vertex = mesh->CollapseEdge(edge, edata->position);
    if (!vertex) continue;

     // Update vertex quadric
    VertexData *vdata = (VertexData *) mesh->VertexData(vertex);
    assert(vdata && (vdata->vertex == vertex));
    vdata->quadric = edata->quadric;

    // Update edge quadrics in a two-ring neighborhood
    R3mesh_mark++;
    for (int j = 0; j < mesh->VertexValence(vertex); j++) {
      R3MeshEdge *e = mesh->EdgeOnVertex(vertex, j);
      if (mesh->EdgeMark(e) != R3mesh_mark) {
        mesh->SetEdgeMark(e, R3mesh_mark);
        ComputeEdgeQuadric(mesh, e, TRUE);
        EdgeData *ed = (EdgeData *) mesh->EdgeData(e);
        assert(ed);
        heap.Update(ed);
      }
      R3MeshVertex *ev = mesh->VertexAcrossEdge(e, vertex);
      for (int k = 0; k < mesh->VertexValence(ev); k++) {
        R3MeshEdge *eve = mesh->EdgeOnVertex(ev, k);
        if (mesh->EdgeMark(eve) != R3mesh_mark) {
          mesh->SetEdgeMark(eve, R3mesh_mark);
          ComputeEdgeQuadric(mesh, eve, FALSE);
          EdgeData *eved = (EdgeData *) mesh->EdgeData(eve);
          assert(eved);
          heap.Update(eved);
        }
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Simplified mesh ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Faces = %d\n", mesh->NFaces());
    printf("  # Edges = %d\n", mesh->NEdges());
    printf("  # Vertices = %d\n", mesh->NVertices());
    fflush(stdout);
  }

  // Check if mesh is valid
  assert(mesh->IsValid());

  // Return success
  return 1;
}



static int 
SimplifyMesh(R3Mesh *mesh, int nfaces, 
  RNArray<VertexCorrespondence *>& vertex_correspondences,
  RNArray<EdgeCorrespondence *>& edge_correspondences)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Compute quadric for every vertex
  for (int i = 0; i < mesh->NVertices(); i++) {
    R3MeshVertex *vertex = mesh->Vertex(i);
    if (!ComputeVertexQuadric(mesh, vertex)) return 0;
  }

  // Compute quadric for every edge
  for (int i = 0; i < mesh->NEdges(); i++) {
    R3MeshEdge *edge = mesh->Edge(i);
    if (!ComputeEdgeQuadric(mesh, edge, TRUE)) return 0;
  }

  // Compute error for every edge correspondence
  for (int i = 0; i < edge_correspondences.NEntries(); i++) {
    EdgeCorrespondence *correspondence = edge_correspondences.Kth(i);
    ComputeEdgeCorrespondenceError(mesh, correspondence);
  }

  // Build priority queue
  EdgeCorrespondence tmp;
  RNHeap<EdgeCorrespondence *> heap(&tmp, &tmp.error, &tmp.heapentry);
  for (int i = 0; i < edge_correspondences.NEntries(); i++) {
    EdgeCorrespondence *correspondence = edge_correspondences.Kth(i);
    heap.Push(correspondence);
  }

  // Collapse edge correspondences
  while (mesh->NFaces() > nfaces) {
    EdgeCorrespondence *correspondence = heap.Pop();
    if (!correspondence) break;

#if 0
    // Check correspondence error
    RNScalar error = correspondence->error;
    for (int i = 0; i < correspondence->edges.NEntries(); i++)
      ComputeEdgeQuadric(mesh, correspondence->edges.Kth(i), TRUE);
    ComputeEdgeCorrespondenceError(mesh, correspondence);
    if (correspondence->error != error) {
      printf("Mismatched error: %g %g %d\n", error, correspondence->error, mesh->NFaces());
      heap.Push(correspondence);
      continue;
    }
#endif

    // Check if we are done
    if (correspondence->error >= RN_INFINITY) break;

    // Print a message every once in a while
    if (print_verbose) {
      static int next_nfaces = 0;
      if (next_nfaces == 0) {
        next_nfaces = 1000 * (mesh->NFaces() / 1000);
      }
      else if (mesh->NFaces() < next_nfaces) { 
        RNArea area = 0;
        for (int i = 0; i < correspondence->edges.NEntries(); i++) {
          R3MeshEdge *edge = correspondence->edges.Kth(i);
          R3MeshFace *f0 = mesh->FaceOnEdge(edge, 0);
          if (f0) area += mesh->FaceArea(f0);
          R3MeshFace *f1 = mesh->FaceOnEdge(edge, 1);
          if (f1) area += mesh->FaceArea(f1);
        }
        printf("%d faces, %g error, %g area\n", mesh->NFaces(), correspondence->error, area);
        next_nfaces -= 1000;
      }
    }

    // Collapse all edges in correspondence
    RNArray<R3MeshEdge *> corresponding_edges = correspondence->edges;
    for (int i = 0; i < corresponding_edges.NEntries(); i++) {
      R3MeshEdge *edge = corresponding_edges.Kth(i);
      EdgeData *edata = (EdgeData *) mesh->EdgeData(edge);
      assert(edata && edata->correspondence && (edata->correspondence == correspondence));

      // Remember data of what will be deleted
      R3MeshVertex *v1 = mesh->VertexOnEdge(edge, 1);
      R3MeshFace *f0 = mesh->FaceOnEdge(edge, 0);  
      R3MeshFace *f1 = mesh->FaceOnEdge(edge, 1);  
      R3MeshEdge *e01 = (f0) ? mesh->EdgeAcrossVertex(v1, edge, f0) : NULL;
      R3MeshEdge *e11 = (f1) ? mesh->EdgeAcrossVertex(v1, edge, f1) : NULL;
      VertexData *vdata1 = (VertexData *) mesh->VertexData(v1);
      EdgeData *edata01 = (e01) ? (EdgeData *) mesh->EdgeData(e01) : NULL;
      EdgeData *edata11 = (e11) ? (EdgeData *) mesh->EdgeData(e11) : NULL;

      // Collapse edge
      R3MeshVertex *vertex = mesh->CollapseEdge(edge, edata->position);
      if (!vertex) continue;

      // Update vertex correspondences (Removed v1)
      if (vdata1 && vdata1->correspondence) {
        vdata1->correspondence->vertices.Remove(v1);
        vdata1->correspondence = NULL;
      }

      // Update edge correspondences (Removed edge, e01, and e11)
      correspondence->edges.RemoveKth(i);
      if (edata01 && edata01->correspondence) {
        edata01->correspondence->edges.Remove(e01);
        edata01->correspondence = NULL;
      }
      if (edata11 && edata11->correspondence) {
        edata11->correspondence->edges.Remove(e11);
        edata11->correspondence = NULL;
      }
 
      // Update vertex quadric
      VertexData *vdata = (VertexData *) mesh->VertexData(vertex);
      assert(vdata && (vdata->vertex == vertex));
      vdata->quadric = edata->quadric;

      // Update edge quadrics in a two-ring neighborhood
      R3mesh_mark++;
      for (int j = 0; j < mesh->VertexValence(vertex); j++) {
        R3MeshEdge *e = mesh->EdgeOnVertex(vertex, j);
        if (mesh->EdgeMark(e) != R3mesh_mark) {
          mesh->SetEdgeMark(e, R3mesh_mark);
          ComputeEdgeQuadric(mesh, e, TRUE);
          EdgeData *ed = (EdgeData *) mesh->EdgeData(e);
          assert(ed);
          if (ed->correspondence) {
            ComputeEdgeCorrespondenceError(mesh, ed->correspondence);
            heap.Update(ed->correspondence);
          }
        }
        R3MeshVertex *ev = mesh->VertexAcrossEdge(e, vertex);
        for (int k = 0; k < mesh->VertexValence(ev); k++) {
          R3MeshEdge *eve = mesh->EdgeOnVertex(ev, k);
          if (mesh->EdgeMark(eve) != R3mesh_mark) {
            mesh->SetEdgeMark(eve, R3mesh_mark);
            ComputeEdgeQuadric(mesh, eve, FALSE);
            EdgeData *eved = (EdgeData *) mesh->EdgeData(eve);
            assert(eved);
            if (eved->correspondence) {
              ComputeEdgeCorrespondenceError(mesh, eved->correspondence);
              heap.Update(eved->correspondence);
            }
          }
        }
      }
    }
  }

  // Update vertex correspondences
  for (int i = vertex_correspondences.NEntries()-1; i >= 0; i--) {
    VertexCorrespondence *correspondence = vertex_correspondences.Kth(i);
    if (correspondence->vertices.IsEmpty()) {
      RNArrayEntry *entry = vertex_correspondences.KthEntry(i);
      vertex_correspondences.EntryContents(entry) = vertex_correspondences.Tail();
      vertex_correspondences.Truncate(vertex_correspondences.NEntries()-1);
    }
  }

  // Update edge correspondences
  for (int i = edge_correspondences.NEntries()-1; i >= 0; i--) {
    EdgeCorrespondence *correspondence = edge_correspondences.Kth(i);
    if (correspondence->edges.IsEmpty()) {
      RNArrayEntry *entry = edge_correspondences.KthEntry(i);
      edge_correspondences.EntryContents(entry) = edge_correspondences.Tail();
      edge_correspondences.Truncate(edge_correspondences.NEntries()-1);
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Simplified mesh ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Faces = %d\n", mesh->NFaces());
    printf("  # Edges = %d\n", mesh->NEdges());
    printf("  # Vertices = %d\n", mesh->NVertices());
    fflush(stdout);
  }

  // Check if mesh is valid
  assert(mesh->IsValid());

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// USER INTERFACE
////////////////////////////////////////////////////////////////////////

void drawPlane2(R3Point c, R3Vector d1, R3Vector d2, R3Vector normal, R3Point color){
	glDisable(GL_LIGHTING);
	glLineWidth(2);
	glColor3f(color[0], color[1], color[2]);
	int numLines = 13;
	for(int i = 0; i < numLines; i++){
		glBegin(GL_LINES);
		R3Vector step = d2 * 2 / (numLines - 1);
		R3LoadPoint(c - d1 + step * i - d2);
		R3LoadPoint(c + d1 + step * i - d2);
		glEnd();	
	}

	for(int i = 0; i < numLines; i++){
		glBegin(GL_LINES);
		R3Vector step = d1 * 2 / (numLines - 1);
		R3LoadPoint(c - d2 + step * i - d1);
		R3LoadPoint(c + d2 + step * i - d1);
		glEnd();	
	}
}

void drawPlane(R3Point c, R3Vector d1, R3Vector d2, R3Vector normal, R3Point color){
	glDisable(GL_LIGHTING);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDepthMask(FALSE);
	glEnable(GL_BLEND);
	glColor4f(color[0], color[1], color[2], 0.2);
	glBegin(GL_POLYGON);
	R3LoadNormal(normal);
	R3LoadPoint(c + d1 + d2);
	R3LoadPoint(c + d1 - d2);
	R3LoadPoint(c - d1 - d2);
	R3LoadPoint(c - d1 + d2);
	glEnd();
	glDisable(GL_BLEND);
	glDepthMask(TRUE);

}




void GLUIStop(void)
{
  // Destroy window 
  glutDestroyWindow(GLUIwindow);

  // Exit
  exit(0);
}



void GLUIIdle(void)
{
  // Set current window
  //  if ( glutGetWindow() != GLUIwindow ) 
  //    glutSetWindow(GLUIwindow);  

  // Redraw
  glutPostRedisplay();
}



void GLUIDrawText(const R3Point& p, const char *s)
{
  // Draw text string s and position p
  glRasterPos3d(p[0], p[1], p[2]);
  while (*s) glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, *(s++));
}
  


void GLUIRedraw(void)
{
  // Set viewing transformation
  viewer->Camera().Load();

  // Clear window 
  glClearColor(1.0, 1.0, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Set backface culling
  if (show_backfacing) glDisable(GL_CULL_FACE);
  else glEnable(GL_CULL_FACE);

  // Handle rocking
  if (rocking) {
    static int frame_count = 0;
    static int frame_max = 90;
    static RNAngle frame_angle = RN_PI / 180;
    int frame = (frame_count % frame_max);
    if (frame > frame_max/2) frame = frame_max - frame;
    RNAngle angle = -1 * frame_max/4 * frame_angle + frame * frame_angle;
    RNAngle tmp = pow(fabs(angle), .85);
    angle = (angle >= 0) ? tmp : -tmp;
    R3Vector center = mesh->BBox().Centroid().Vector();
    R4Matrix matrix = R4identity_matrix;
    matrix.Translate(center);
    matrix.Rotate(viewer->Camera().Up(), angle);
    matrix.Translate(-center);
    matrix.Draw();
    frame_count++;
    if (output_image_name) {
      if (frame_count <= frame_max) {
        static int first = 1;
        static char image_name[256];
        static char dir_name[256];
        if (first) {
          char buffer[256];
          strcpy(dir_name, output_image_name);
          sprintf(buffer, "mkdir %s", dir_name);
          system(buffer);
          first = 0;
        }
        sprintf(image_name, "%s/%04d.jpg", dir_name, frame_count);
        output_image_name = image_name;
        save_image = 1;
      }
      else {
        output_image_name = NULL;
        rocking = 0;
      }
    }
  }

#ifndef ALEX_LIGHTS
  // Set lights
  static GLfloat light0_position[] = { 3.0, 4.0, 5.0, 0.0 };
  glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
  static GLfloat light1_position[] = { -3.0, -2.0, -3.0, 0.0 };
  glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
#endif

  // Draw faces
  if (show_faces) {
    glEnable(GL_LIGHTING);
    GLfloat diffuse_material[] = { 1.0, 0.5, 0.5, 1.0 };
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, diffuse_material); 
    // GLfloat specular_material[] = { 0.1, 0.1, 0.1, 1.0 };
    // glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular_material); 
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(2, 1);
    glBegin(GL_TRIANGLES);
    for (int i = 0; i < mesh->NFaces(); i++) {
      R3MeshFace *face = mesh->Face(i);
      const R3Vector& n = mesh->FaceNormal(face);
      if (!show_gouraud) glNormal3d(n[0], n[1], n[2]);
      for (int j = 0; j < 3; j++) {
        R3MeshVertex *vertex = mesh->VertexOnFace(face, j);
        const R3Point& p = mesh->VertexPosition(vertex);
        const R3Vector& n = mesh->VertexNormal(vertex);
        if (show_gouraud) glNormal3d(n[0], n[1], n[2]);
        glVertex3d(p[0], p[1], p[2]);
      }
    }
    glEnd();
    glDisable(GL_POLYGON_OFFSET_FILL);
  }
   
  // Draw reflected faces
  if (!planes.IsEmpty() && show_reflected_faces) {
    glEnable(GL_LIGHTING);
    static GLfloat material[] = { 0.0, 1.0, 0.0, 1.0 };
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, material); 
    R3Affine affine(R3identity_affine);
    affine.Mirror(*(planes[0]));
    for (int i = 0; i < mesh->NFaces(); i++) {
      R3MeshFace *face = mesh->Face(i);
      R3BeginPolygon();
      R3Vector normal = mesh->FaceNormal(face);
      normal.Transform(affine);
      glNormal3d(normal[0], normal[1], normal[2]);
      for (int j = 2; j >= 0; j--) {
        R3MeshVertex *vertex = mesh->VertexOnFace(face, j);
        R3Point position = mesh->VertexPosition(vertex);
        position.Transform(affine);
        R3LoadPoint(position);
      }
      R3EndPolygon();
    }
  }

  // Draw edges
  if (show_edges) {
    glDisable(GL_LIGHTING);
    glColor3f(0, 0, 0);
    glLineWidth(3);
    mesh->DrawEdges();

    // Draw edge names
    if (show_names) {
      static char buffer[256];
      glDisable(GL_LIGHTING);
      glColor3f(0, 0, 0);
      for (int i = 0; i < mesh->NEdges(); i++) {
        R3MeshEdge *edge = mesh->Edge(i);
        R3Point position = mesh->EdgeMidpoint(edge);
        sprintf(buffer, "%d", mesh->EdgeID(edge));
        GLUIDrawText(position, buffer);
      }
    }
    glLineWidth(1);
  }

  // Draw vertices
  if (show_vertices) {
    glEnable(GL_LIGHTING);
    static GLfloat material[] = { 0.8, 0.4, 0.2, 1.0 };
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material); 
    RNScalar d = 0.001 * mesh->BBox().LongestAxisLength();
    for (int i = 0; i < mesh->NVertices(); i++) {
      R3MeshVertex *vertex = mesh->Vertex(i);
      R3Point position = mesh->VertexPosition(vertex);
      R3Sphere(position, d).Draw();
    }

    // Draw vertex names
    if (show_names) {
      static char buffer[256];
      glDisable(GL_LIGHTING);
      glColor3f(0, 0.8, 0);
      for (int i = 0; i < mesh->NVertices(); i++) {
        R3MeshVertex *vertex = mesh->Vertex(i);
        R3Point position = mesh->VertexPosition(vertex);
        sprintf(buffer, "%d", mesh->VertexID(vertex));
        GLUIDrawText(position, buffer);
      }
    }
  }

  // Draw edge errors
  if (show_errors) {
    glDisable(GL_LIGHTING);
    for (int i = 0; i < mesh->NEdges(); i++) {
      R3MeshEdge *edge = mesh->Edge(i);
      EdgeData *edata = (EdgeData *) mesh->EdgeData(edge);
      GLdouble value = edata->error * show_errors_scale;
      if (edata->error >= RN_INFINITY) glColor3f(1, 0.5, 0);
      else glColor3f(value, value, value);
      mesh->DrawEdge(edge);
      char buffer[256];
      sprintf(buffer, "%3g", edata->error);
      GLUIDrawText(mesh->EdgeMidpoint(edge), buffer);
    }
  }

  // Draw vertex correspondences
  if (show_vertex_correspondences && vertex_correspondences) {
    glEnable(GL_LIGHTING);
    for (int i = 0; i < vertex_correspondences->NEntries(); i++) {
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, colors[i%ncolors]); 
      VertexCorrespondence *vertex_correspondence = vertex_correspondences->Kth(i);
      RNScalar d = 0.003 * mesh->BBox().LongestAxisLength();
      for (int j = 0; j < vertex_correspondence->vertices.NEntries(); j++) {
        R3MeshVertex *vertex = vertex_correspondence->vertices.Kth(j);
        R3Point position = mesh->VertexPosition(vertex);
        R3Sphere(position, d).Draw();
      }
    }
  }

  // Draw edge correspondences
  if (show_edge_correspondences && edge_correspondences) {
    glDisable(GL_LIGHTING);
    for (int i = 0; i < edge_correspondences->NEntries(); i++) {
      glColor4fv(colors[i%ncolors]); 
      EdgeCorrespondence *edge_correspondence = edge_correspondences->Kth(i);
      for (int j = 0; j < edge_correspondence->edges.NEntries(); j++) {
        R3MeshEdge *edge = edge_correspondence->edges.Kth(j);
        mesh->DrawEdge(edge);
      }
    }
  }

  // Draw face correspondences
  if (show_face_correspondences && face_correspondences) {
    glEnable(GL_LIGHTING);
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(-1, 1);
    for (int i = 0; i < face_correspondences->NEntries(); i++) {
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, colors[i%ncolors]); 
      FaceCorrespondence *face_correspondence = face_correspondences->Kth(i);
      for (int j = 0; j < face_correspondence->faces.NEntries(); j++) {
        R3MeshFace *face = face_correspondence->faces.Kth(j);
        mesh->DrawFace(face);
      }
    }
    glDisable(GL_POLYGON_OFFSET_FILL);
  }

  // Draw planes
  if (show_planes) {
#if 1
    glDisable(GL_CULL_FACE);
    for (int i = 0; i < planes.NEntries(); i++) {
      R3Plane *plane = planes.Kth(i);
      // Compute points for current plane
      R3Point center = mesh->BBox().Centroid();
      R3Point c = center + (plane->Point() - center).Dot(plane->Normal()) * plane->Normal();
      R3Vector d1 = plane->Normal() % R3Vector(1, 0, 0);
      if(d1.Length() < .1) d1 = plane->Normal() % R3Vector(0, 1, 0);
      d1.Normalize();
      d1 *=  0.75 * mesh->BBox().DiagonalRadius();
      R3Vector d2 = plane->Normal() % d1;
      // Draw current plane
      drawPlane(c, d1, d2, plane->Normal(), R3Point(.5, .5, .5));
      drawPlane2(c, d1, d2, plane->Normal(), R3Point(0, 1, 0));
    }
    glEnable(GL_CULL_FACE);
#else
    glEnable(GL_LIGHTING);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    static GLfloat material[] = { 0.8, 0.8, 0.8, 0.5 };
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
    for (int i = 0; i < planes.NEntries(); i++) {
      R3Plane *plane = planes.Kth(i);
      R3Point center = mesh->BBox().Centroid();
      center.Project(*plane);
      R3Vector towards = plane->Normal();
      R3Vector up = R3xyz_triad[towards.MinDimension()];
      R3Triad triad(towards, up);
      R3CoordSystem cs(center, triad); 
      R4Matrix m = cs.Matrix();
      RNScalar r = mesh->BBox().DiagonalRadius();
      m.Push();
      R3Point p(r, r, 0);
      R3Box(-p, p).Draw();
      m.Pop();
    }
    glDisable(GL_BLEND);
#endif
  }

  // Draw axes
  if (show_axes) {
    RNScalar d = mesh->BBox().DiagonalRadius();
    glDisable(GL_LIGHTING);
    glLineWidth(3);
    R3BeginLine();
    glColor3f(1, 0, 0);
    R3LoadPoint(R3zero_point + d * R3negx_vector);
    R3LoadPoint(R3zero_point + d * R3posx_vector);
    R3EndLine();
    R3BeginLine();
    glColor3f(0, 1, 0);
    R3LoadPoint(R3zero_point + d * R3negy_vector);
    R3LoadPoint(R3zero_point + d * R3posy_vector);
    R3EndLine();
    R3BeginLine();
    glColor3f(0, 0, 1);
    R3LoadPoint(R3zero_point + d * R3negz_vector);
    R3LoadPoint(R3zero_point + d * R3posz_vector);
    R3EndLine();
    glLineWidth(1);
  }

  // Save image
  if (save_image) {
    char buffer[256];
    static int image_number = 1;
    if (output_image_name) strcpy(buffer, output_image_name);
    else sprintf(buffer, "image%d.jpg", image_number++);
    R2Image image(GLUIwindow_width, GLUIwindow_height, 3);
    image.Capture();
    if (image.Write(buffer)) printf("Saved %s\n", buffer);
    if (!interactive) GLUIStop();
    save_image = 0;
  }

  // Swap buffers 
  glutSwapBuffers();
}    



void GLUIResize(int a, int b)
{
  // Resize window
  int x, y, w, h;
  GLUI_Master.get_viewport_area( &x, &y, &w, &h );
  glViewport( x, y, w, h );

  // Resize viewer viewport
  viewer->ResizeViewport(x, y, w, h);

  // Remember window size 
  GLUIwindow_width = w;
  GLUIwindow_height = h;

  // Redraw
  glutPostRedisplay();
}



void GLUIMotion(int x, int y)
{
  // Invert y coordinate
  y = GLUIwindow_height - y;

  // Compute mouse movement
  int dx = x - GLUImouse[0];
  int dy = y - GLUImouse[1];
  
  // World in hand navigation 
  R3Point origin = mesh->BBox().Centroid();
  if (GLUIbutton[0]) viewer->RotateWorld(1.0, origin, x, y, dx, dy);
  else if (GLUIbutton[1]) viewer->ScaleWorld(1.0, origin, x, y, dx, dy);
  else if (GLUIbutton[2]) viewer->TranslateWorld(1.0, origin, x, y, dx, dy);
  if (GLUIbutton[0] || GLUIbutton[1] || GLUIbutton[2]) glutPostRedisplay();

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = y;
}



void GLUIMouse(int button, int state, int x, int y)
{
  // Invert y coordinate
  y = GLUIwindow_height - y;
  
  // Process mouse button event

  // Remember button state 
  int b = (button == GLUT_LEFT_BUTTON) ? 0 : ((button == GLUT_MIDDLE_BUTTON) ? 1 : 2);
  GLUIbutton[b] = (state == GLUT_DOWN) ? 1 : 0;

  // Remember modifiers 
  GLUImodifiers = glutGetModifiers();

   // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = y;

  // Redraw
  glutPostRedisplay();
}



void GLUISpecial(int key, int x, int y)
{
  // Invert y coordinate
  y = GLUIwindow_height - y;

  // Process keyboard button event 

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = y;

  // Remember modifiers 
  GLUImodifiers = glutGetModifiers();

  // Redraw
  glutPostRedisplay();
}



void GLUIKeyboard(unsigned char key, int x, int y)
{
  // Process keyboard button event 
  switch (key) {
  case '~': 
    save_image = 1;
    break;

  case 'A':
  case 'a':
    show_axes = !show_axes;
    break;

  case 'B':
  case 'b':
    show_backfacing = !show_backfacing;
    break;

  case 'C': 
  case 'c': {
    // Print camera
    const R3Camera& camera = viewer->Camera();
    printf("-window %d %d\n", GLUIwindow_width, GLUIwindow_height);
    printf("-camera  %g %g %g  %g %g %g  %g %g %g\n",
           camera.Origin().X(), camera.Origin().Y(), camera.Origin().Z(),
           camera.Towards().X(), camera.Towards().Y(), camera.Towards().Z(),
           camera.Up().X(), camera.Up().Y(), camera.Up().Z());
    }
    break;
      
  case 'E':
    show_edge_correspondences = !show_edge_correspondences;
    break;

  case 'e':
    show_edges = !show_edges;
    break;

  case 'F':
    show_face_correspondences = !show_face_correspondences;
    break;

  case 'f':
    show_faces = !show_faces;
    break;

  case 'G':
  case 'g':
    show_gouraud = !show_gouraud;
    break;

  case 'M':
  case 'm':
    show_errors = !show_errors;
    break;

  case 'N':
  case 'n':
    show_names = !show_names;
    break;

  case 'P':
  case 'p':
    show_planes = !show_planes;
    break;

  case 'R':
  case 'r':
    show_reflected_faces = !show_reflected_faces;
    break;

  case 'S':
    show_errors_scale *= 2;
    break;

  case 's':
    show_errors_scale /= 2;
    break;

  case 'V':
    show_vertex_correspondences = !show_vertex_correspondences;
    break;

  case 'v':
    show_vertices = !show_vertices;
    break;

  case 27: // ESCAPE
    GLUIStop();
    break;
  }

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = GLUIwindow_height - y;

  // Remember modifiers 
  GLUImodifiers = glutGetModifiers();

  // Synchronize GLUI interface
  GLUI_Master.sync_live_all();

  // Redraw
  glutPostRedisplay();  
}




void GLUICallback(int control)
{
  switch (control) {
  case QUIT_COMMAND:
    GLUIStop();
    break;
  }
}



void GLUISetup(void)
{
  // Initialize GLUI
  glui = GLUI_Master.create_glui("OpenGL Viewer Controls", 0, GLUIwindow_width + 120, 100);
  glui->set_main_gfx_window(GLUIwindow);

  // Add check box
  GLUI_Panel *panel = glui->add_panel("Display");
  glui->add_checkbox_to_panel(panel, "Vertices (V)", &show_vertices );
  glui->add_checkbox_to_panel(panel, "Edges    (E)", &show_edges );
  glui->add_checkbox_to_panel(panel, "Faces    (F)", &show_faces );

  // Add button
  glui->add_button("Quit", QUIT_COMMAND, GLUICallback);
}



void GLUIInit(int *argc, char **argv)
{
  // Open window 
  glutInit(argc, argv);
  glutInitWindowPosition(100, 100);
  glutInitWindowSize(GLUIwindow_width, GLUIwindow_height);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH); // | GLUT_STENCIL
  GLUIwindow = glutCreateWindow("OpenGL Viewer");

  // Initialize background color 
  glClearColor(200.0/255.0, 200.0/255.0, 200.0/255.0, 1.0);

#ifndef ALEX_LIGHTS
  // Initialize lights
  static GLfloat lmodel_ambient[] = { 0.2, 0.2, 0.2, 1.0 };
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  static GLfloat light0_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
  glEnable(GL_LIGHT0);
  static GLfloat light1_diffuse[] = { 0.5, 0.5, 0.5, 1.0 };
  glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
  glEnable(GL_LIGHT1);
  glEnable(GL_NORMALIZE);
  glEnable(GL_LIGHTING);
#else
  // Setup lighting (from Alex)
  GLfloat mat_specular[4] = { 0.18, 0.18, 0.18, 0.18 };
  GLfloat mat_shininess[] = { 64 };
  GLfloat global_ambient[] = { 0.02, 0.02, 0.05, 0.05 };
  GLfloat light0_ambient[] = { 0, 0, 0, 0 };
  GLfloat light0_diffuse[] = { 0.85, 0.85, 0.8, 0.85 };
  GLfloat light1_diffuse[] = { -0.01, -0.01, -0.03, -0.03 };
  GLfloat light0_specular[] = { 0.85, 0.85, 0.85, 0.85 };
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);
  glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light0_specular);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  glEnable(GL_NORMALIZE);
#endif

  // Initialize graphics modes 
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);

  // Initialize GLUT callback functions 
  glutDisplayFunc(GLUIRedraw);
  GLUI_Master.set_glutReshapeFunc(GLUIResize);
  GLUI_Master.set_glutKeyboardFunc(GLUIKeyboard);
  GLUI_Master.set_glutSpecialFunc(GLUISpecial);
  GLUI_Master.set_glutMouseFunc(GLUIMouse);
  glutMotionFunc(GLUIMotion);
  if (rocking)GLUI_Master.set_glutIdleFunc(GLUIIdle);

  // Setup GLUI controls
  // GLUISetup();
}



void GLUIMainLoop(void)
{
  // Run main loop -- never returns 
  glutMainLoop();
}



static R3Viewer *
CreateBirdsEyeViewer(R3Mesh *mesh)
{
  // Setup camera view looking down the Z axis
  R3Box bbox = mesh->BBox();
  assert(!bbox.IsEmpty());
  RNLength r = bbox.DiagonalRadius();
  assert((r > 0.0) && RNIsFinite(r));
  if (!initial_camera) initial_camera_origin = bbox.Centroid() - initial_camera_towards * (5 * r);
  R3Camera camera(initial_camera_origin, initial_camera_towards, initial_camera_up, 0.2, 0.2, 0.1 * r, 10.0 * r);
  R2Viewport viewport(0, 0, GLUIwindow_width, GLUIwindow_height);
  return new R3Viewer(camera, viewport);
}



////////////////////////////////////////////////////////////////////////
// I/O STUFF
////////////////////////////////////////////////////////////////////////

static R3Mesh *
ReadMesh(char *mesh_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate mesh
  R3Mesh *mesh = new R3Mesh();
  assert(mesh);

  // Read mesh from file
  if (!mesh->ReadFile(mesh_name)) {
    delete mesh;
    return NULL;
  }

  // Check if mesh is valid
  assert(mesh->IsValid());

  // Set scale
  scale = mesh->BBox().DiagonalLength();
  RN_EPSILON = 1.0E-6 * scale;
  RN_INFINITY = 1.0E6 * scale;

  // Print statistics
  if (print_verbose) {
    printf("Read mesh ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Faces = %d\n", mesh->NFaces());
    printf("  # Edges = %d\n", mesh->NEdges());
    printf("  # Vertices = %d\n", mesh->NVertices());
    printf("  Scale = %g\n", scale);
    fflush(stdout);
  }

  // Return success
  return mesh;
}



static int
WriteMesh(R3Mesh *mesh, char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write mesh to file
  if (!mesh->WriteFile(filename)) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Wrote mesh to %s ...\n", filename);
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Faces = %d\n", mesh->NFaces());
    printf("  # Edges = %d\n", mesh->NEdges());
    printf("  # Vertices = %d\n", mesh->NVertices());
    fflush(stdout);
  }

  // Return success
  return 1;
}



static RNArray<VertexCorrespondence *> * 
ReadVertexCorrespondences(R3Mesh *mesh, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = fopen(filename, "r");
  if (!fp) {
    fprintf(stderr, "Unable to open correspondences file: %s\n", filename);
    return 0;
  }

  // Allocate vertex data
  for (int i = 0; i < mesh->NVertices(); i++) {
    R3MeshVertex *vertex = mesh->Vertex(i);
    VertexData *vdata = new VertexData();
    vdata->vertex = vertex;
    vdata->quadric = R4null_matrix;
    vdata->correspondence = NULL;
    mesh->SetVertexData(vertex, vdata);
  }

  // Allocate array of vertex correspondences
  RNArray<VertexCorrespondence *> *vertex_correspondences = new RNArray<VertexCorrespondence *>();
  if (!vertex_correspondences) {
    fprintf(stderr, "Unable to allocate vertex correspondences\n");
    return NULL;
  }

  // Read number of vertex correspondences
  int ncorrespondences;
  if (fscanf(fp, "%d", &ncorrespondences) != 1) {
    fprintf(stderr, "Unable to read vertex correspondences from %s\n", filename);
    return 0; 
  }

  // Read vertex correspondences
  for (int i = 0; i < ncorrespondences; i++) {
    // Create correspondnce
    VertexCorrespondence *correspondence = new VertexCorrespondence();
    vertex_correspondences->Insert(correspondence);

    // Read number of vertices in correspondence
    int nvertices;
    if (fscanf(fp, "%d", &nvertices) != 1) {
      fprintf(stderr, "Unable to read number of vertices in correspondence %d from %s\n", i, filename);
      return 0; 
    }

    // Read vertex IDs
    for (int j = 0; j < nvertices; j++) {
      // Read vertex index
      int vertex_index;
      if (fscanf(fp, "%d", &vertex_index) != 1) {
        fprintf(stderr, "Unable to read vertex id %d in correspondence %d from %s\n", j, i, filename);
        return 0;
      }

      // Get vertex
      R3MeshVertex *vertex = mesh->Vertex(vertex_index);
      VertexData *vdata = (VertexData *) mesh->VertexData(vertex);
      assert(vdata);

      // Insert vertex into correspondence, and vice-versa
      correspondence->vertices.Insert(vertex);
      vdata->correspondence = correspondence;
    }
  }

  // Close file
  fclose (fp);

  // Print statistics
  if (print_verbose) {
    // Count number of matched vertices
    int matched = 0;
    for (int i = 0; i < mesh->NVertices(); i++) {
      R3MeshVertex *vertex = mesh->Vertex(i);
      VertexData *vdata = (VertexData *) mesh->VertexData(vertex);
      if (!vdata || !vdata->correspondence) continue;
      matched++;
    }
 
    // Count number of multi-vertex correspondences
    int multi = 0;
    for (int i = 0; i < vertex_correspondences->NEntries(); i++) {
      VertexCorrespondence *correspondence = vertex_correspondences->Kth(i);
      if (correspondence->vertices.NEntries() <= 2) continue;
      multi++;
    }

    // Print message 
    printf("Read correspondences from %s ...\n", filename);
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Vertices = %d\n", mesh->NVertices());
    printf("  # Vertex correspondences = %d\n", vertex_correspondences->NEntries());
    printf("  # Multiple Vertex Correspondences = %d\n", multi);
    printf("  # Vertices Matched = %d\n", matched);
    printf("  # Vertices Unmatched = %d\n", mesh->NVertices() - matched);
    fflush(stdout);
  }

  // Return vertex correspondences
  return vertex_correspondences;
}



static RNArray<EdgeCorrespondence *> *
CreateEdgeCorrespondences(R3Mesh *mesh, 
  RNArray<VertexCorrespondence *>& vertex_correspondences)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate edge data
  for (int i = 0; i < mesh->NEdges(); i++) {
    R3MeshEdge *edge = mesh->Edge(i);
    EdgeData *edata = (EdgeData *) mesh->EdgeData(edge);
    if (!edata) {
      edata = new EdgeData();
      edata->edge = edge;
      edata->quadric = R4null_matrix;
      edata->quadric_error = 0;
      edata->error = 0;
      edata->position = R3zero_point;
      edata->id = mesh->EdgeID(edge);
      edata->heapentry = NULL;
      edata->correspondence = NULL;
      mesh->SetEdgeData(edge, edata);
    }
  }

  // Allocate array of edge correspondences
  RNArray<EdgeCorrespondence *> *edge_correspondences = new RNArray<EdgeCorrespondence *>();
  if (!edge_correspondences) {
    fprintf(stderr, "Unable to allocate array of edge correspondences.\n");
    return NULL;
  }

  // Check all edges for correspondences
  for (int i = 0; i < mesh->NEdges(); i++) {
    R3MeshEdge *edge = mesh->Edge(i);
    EdgeData *edata = (EdgeData *) mesh->EdgeData(edge);
    assert(edata);

    // Check if already has correspondences
    if (edata->correspondence) continue;

    // Get vertex correspondences
    R3MeshVertex *v0 = mesh->VertexOnEdge(edge, 0);
    R3MeshVertex *v1 = mesh->VertexOnEdge(edge, 1);
    assert(v0 && v1);
    VertexData *v0_data = (VertexData *) mesh->VertexData(v0);
    VertexData *v1_data = (VertexData *) mesh->VertexData(v1);
    assert(v0_data && v1_data);
    VertexCorrespondence *v0_correspondence = v0_data->correspondence;
    VertexCorrespondence *v1_correspondence = v1_data->correspondence;
    if (!v0_correspondence || !v1_correspondence) continue;

    // Find edge correspondences
    for (int j = 0; j < v0_correspondence->vertices.NEntries(); j++) {
      R3MeshVertex *v0_v = v0_correspondence->vertices[j];
      for (int k = 0; k < v1_correspondence->vertices.NEntries(); k++) {
        R3MeshVertex *v1_v = v1_correspondence->vertices[k];
        R3MeshEdge *corresponding_edge = mesh->EdgeBetweenVertices(v0_v, v1_v);
        if (corresponding_edge) {
          if (corresponding_edge == edge) {
            if ((((v0 == v0_v) && (v1 == v1_v)) &&
                 (v0_correspondence->vertices.NEntries() == 1) &&
                 (v1_correspondence->vertices.NEntries() == 1)) ||
                ((v0 == v1_v) && (v1 == v0_v))) {
              // Create edge correspondence with itself
              if (!edata->correspondence) {
                edata->correspondence = new EdgeCorrespondence();
                edata->correspondence->error = 0;
                edata->correspondence->heapentry = 0;
                edata->correspondence->mark = 0;
                edata->correspondence->edges.Insert(edge);
                edge_correspondences->Insert(edata->correspondence);
              }
            }
          }
          else {
            // Check if corresponding edge already has correspondence
            EdgeData *corresponding_edata = (EdgeData *) mesh->EdgeData(corresponding_edge);
            assert(corresponding_edata);
            if (corresponding_edata->correspondence) {
              // Add edge to existing correspondence
              edata->correspondence = corresponding_edata->correspondence;
              edata->correspondence->edges.Insert(edge);
              if (v0_v == mesh->VertexOnEdge(corresponding_edge, 1)) {
                assert(v1_v == mesh->VertexOnEdge(corresponding_edge, 0));
                mesh->FlipEdge(edge);
              }
            }
            else {
              // Create new correspondence for this pair
              edata->correspondence = new EdgeCorrespondence();
              edata->correspondence->error = 0;
              edata->correspondence->heapentry = 0;
              edata->correspondence->mark = 0;
              edata->correspondence->edges.Insert(edge);
              edata->correspondence->edges.Insert(corresponding_edge);
              corresponding_edata->correspondence = edata->correspondence;
              edge_correspondences->Insert(edata->correspondence);
              if (v0_v == mesh->VertexOnEdge(corresponding_edge, 1)) {
                assert(v1_v == mesh->VertexOnEdge(corresponding_edge, 0));
                mesh->FlipEdge(edge);
              }
            }
          }
        }
      }
    }
  }
  
  // Print statistics
  if (print_verbose) {
    // Count number of matched edges
    int matched = 0;
    for (int i = 0; i < mesh->NEdges(); i++) {
      R3MeshEdge *edge = mesh->Edge(i);
      EdgeData *edata = (EdgeData *) mesh->EdgeData(edge);
      if (!edata || !edata->correspondence) continue;
      matched++;
    }
 
    // Count number of multi-edge correspondences
    int multi = 0;
    for (int i = 0; i < edge_correspondences->NEntries(); i++) {
      EdgeCorrespondence *correspondence = edge_correspondences->Kth(i);
      if (correspondence->edges.NEntries() <= 2) continue;
      multi++;
    }

    // Print message 
    printf("Created edge correspondences ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Faces = %d\n", mesh->NFaces());
    printf("  # Edges = %d\n", mesh->NEdges());
    printf("  # Vertices = %d\n", mesh->NVertices());
    printf("  # Edge Correspondences = %d\n", edge_correspondences->NEntries());
    printf("  # Multiple Edge Correspondences = %d\n", multi);
    printf("  # Edges Matched = %d\n", matched);
    printf("  # Edges Unmatched = %d\n", mesh->NEdges() - matched);
    fflush(stdout);
  }

  // Check if mesh is valid
  assert(mesh->IsValid());

  // Return array of edge_correspondences
  return edge_correspondences;
}



static RNArray<FaceCorrespondence *> *
CreateFaceCorrespondences(R3Mesh *mesh, 
  const RNArray<VertexCorrespondence *>& vertex_correspondences,
  const RNArray<EdgeCorrespondence *>& edge_correspondences)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate array of face correspondences
  RNArray<FaceCorrespondence *> *face_correspondences = new RNArray<FaceCorrespondence *>();
  if (!face_correspondences) {
    fprintf(stderr, "Unable to allocate array of face correspondences.\n");
    return NULL;
  }

  // Allocate FaceData structures
  for (int i = 0; i < mesh->NFaces(); i++) {
    R3MeshFace *face = mesh->Face(i);
    FaceData *fdata = new FaceData();
    mesh->SetFaceData(face, fdata);
    fdata->face = face;
    fdata->correspondence = NULL;
  }

  // Check all faces for correspondences
  for (int i = 0; i < mesh->NFaces(); i++) {
    R3MeshFace *face = mesh->Face(i);
    FaceData *fdata = (FaceData *) mesh->FaceData(face);
    assert(fdata);

    // Check if already has correspondences
    if (fdata->correspondence) continue;

    // Get edge correspondences
    R3MeshEdge *e0 = mesh->EdgeOnFace(face, 0);
    R3MeshEdge *e1 = mesh->EdgeOnFace(face, 1);
    R3MeshEdge *e2 = mesh->EdgeOnFace(face, 2);
    assert(e0 && e1 && e2);
    EdgeData *e0_data = (EdgeData *) mesh->EdgeData(e0);
    EdgeData *e1_data = (EdgeData *) mesh->EdgeData(e1);
    EdgeData *e2_data = (EdgeData *) mesh->EdgeData(e2);
    assert(e0_data && e1_data && e2_data);
    EdgeCorrespondence *e0_correspondence = e0_data->correspondence;
    EdgeCorrespondence *e1_correspondence = e1_data->correspondence;
    EdgeCorrespondence *e2_correspondence = e2_data->correspondence;
    if (!e0_correspondence || !e1_correspondence || !e2_correspondence) continue;

    // Find face correspondences
    for (int j = 0; j < e0_correspondence->edges.NEntries(); j++) {
      R3MeshEdge *e0_e = e0_correspondence->edges[j];
      for (int k = 0; k < e1_correspondence->edges.NEntries(); k++) {
        R3MeshEdge *e1_e = e1_correspondence->edges[k];
        R3MeshFace *corresponding_face = mesh->FaceBetweenEdges(e0_e, e1_e);
        if (corresponding_face) {
          for (int m = 0; m < e2_correspondence->edges.NEntries(); m++) {
            R3MeshEdge *e2_e = e2_correspondence->edges[m];
            if (corresponding_face == mesh->FaceBetweenEdges(e1_e, e2_e)) {
              if (corresponding_face == mesh->FaceBetweenEdges(e0_e, e2_e)) {
                if (corresponding_face == face) {
                  if (((e0 == e0_e) && (e1 != e1_e) && (e2 != e2_e)) ||
                      ((e0 != e0_e) && (e1 == e1_e) && (e2 != e2_e)) ||
                      ((e0 != e0_e) && (e1 != e1_e) && (e2 == e2_e))) {
                    // Create face correspondence with itself
                    if (!fdata->correspondence) {
                      fdata->correspondence = new FaceCorrespondence();
                      fdata->correspondence->faces.Insert(face);
                      face_correspondences->Insert(fdata->correspondence);
                    }
                  }
                }
                else {
                  // Check if corresponding face already has correspondence
                  FaceData *corresponding_fdata = (FaceData *) mesh->FaceData(corresponding_face);
                  assert(corresponding_fdata);
                  if (corresponding_fdata->correspondence) {
                    // Don't make multiple way correspondences, for now

                    // Add face to existing correspondence
                    fdata->correspondence = corresponding_fdata->correspondence;
                    fdata->correspondence->faces.Insert(face);
                  }
                  else {
                    // Create new correspondence for this pair
                    fdata->correspondence = new FaceCorrespondence();
                    corresponding_fdata->correspondence = fdata->correspondence;
                    fdata->correspondence->faces.Insert(face);
                    fdata->correspondence->faces.Insert(corresponding_face);
                    face_correspondences->Insert(fdata->correspondence);
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  
  // Print statistics
  if (print_verbose) {
    // Count number of matched faces
    int matched = 0;
    for (int i = 0; i < mesh->NFaces(); i++) {
      R3MeshFace *face = mesh->Face(i);
      FaceData *fdata = (FaceData *) mesh->FaceData(face);
      if (!fdata || !fdata->correspondence) continue;
      matched++;
    }
 
    // Count number of multi-face correspondences
    int multi = 0;
    for (int i = 0; i < face_correspondences->NEntries(); i++) {
      FaceCorrespondence *correspondence = face_correspondences->Kth(i);
      if (correspondence->faces.NEntries() <= 2) continue;
      multi++;
    }

    // Print message 
    printf("Created face correspondences ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Faces = %d\n", mesh->NFaces());
    printf("  # Edges = %d\n", mesh->NEdges());
    printf("  # Vertices = %d\n", mesh->NVertices());
    printf("  # Face Correspondences = %d\n", face_correspondences->NEntries());
    printf("  # Multiple Face Correspondences = %d\n", multi);
    printf("  # Faces Matched = %d\n", matched);
    printf("  # Faces Unmatched = %d\n", mesh->NFaces() - matched);
    fflush(stdout);
  }

  // Check if mesh is valid
  assert(mesh->IsValid());

  // Return array of face_correspondences
  return face_correspondences;
}



static int 
WriteCorrespondences(R3Mesh *mesh, 
  const RNArray<VertexCorrespondence *>& vertex_correspondences, 
  const RNArray<EdgeCorrespondence *>& edge_correspondences, 
  const RNArray<FaceCorrespondence *>& face_correspondences, 
  const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = fopen(filename, "w");
  if (!fp) {
    fprintf(stderr, "Unable to open correspondences file: %s\n", filename);
    return 0;
  }

  // Write vertex correspondences
  fprintf(fp, "%d\n", vertex_correspondences.NEntries());
  for (int i = 0; i < vertex_correspondences.NEntries(); i++) {
    VertexCorrespondence *correspondence = vertex_correspondences[i];
    fprintf(fp, "%d ", correspondence->vertices.NEntries());
    for (int j = 0; j < correspondence->vertices.NEntries(); j++) {
      R3MeshVertex *vertex = correspondence->vertices.Kth(j);
      assert(mesh->Vertex(mesh->VertexID(vertex)) == vertex);
      fprintf(fp, "%d ", mesh->VertexID(vertex));
    }
    fprintf(fp, "\n");
  }

  // Close file
  fclose (fp);

  // Print statistics
  if (print_verbose) {
    printf("Wrote correspondences to %s ...\n", filename);
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Vertex correspondences = %d\n", vertex_correspondences.NEntries());
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// ARGUMENT PARSING STUFF
////////////////////////////////////////////////////////////////////////

static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc == 1) {
    printf("Usage: qslim inputname outputname [options]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { print_verbose = 1; }
      else if (!strcmp(*argv, "-interactive")) { interactive = 1; }
      else if (!strcmp(*argv, "-nfaces")) { argv++; argc--; nfaces = atoi(*argv); }
      else if (!strcmp(*argv, "-correspondences")) { argv++; argc--; input_correspondences_name = *argv; }
      else if (!strcmp(*argv, "-image")) { argv++; argc--; output_image_name = *argv; save_image = 1; }
      else if (!strcmp(*argv, "-outputcorrespondences")) { argv++; argc--; output_correspondences_name = *argv; }
      else if (!strcmp(*argv, "-show_planes")) { show_planes = 1; }
      else if (!strcmp(*argv, "-show_edges")) { show_edges = 1; }
      else if (!strcmp(*argv, "-show_reflected_faces")) { show_reflected_faces = 1; }
      else if (!strcmp(*argv, "-show_vertex_correspondences")) { show_vertex_correspondences = 1; }
      else if (!strcmp(*argv, "-show_edge_correspondences")) { show_edge_correspondences = 1; }
      else if (!strcmp(*argv, "-show_face_correspondences")) { show_face_correspondences = 1; }
      else if (!strcmp(*argv, "-rocking")) { rocking = 1; }
      else if (!strcmp(*argv, "-boundary_weight")) { 
        argv++; argc--; boundary_weight = 0.001 * atof(*argv); 
      }
      else if (!strcmp(*argv, "-window")) { 
        argc--; argv++; GLUIwindow_width = atoi(*argv);
        argc--; argv++; GLUIwindow_height = atoi(*argv);
      }
      else if (!strcmp(*argv, "-plane")) { 
        RNScalar plane_coefficients[4];
        argc--; argv++; plane_coefficients[0] = atof(*argv);
        argc--; argv++; plane_coefficients[1] = atof(*argv);
        argc--; argv++; plane_coefficients[2] = atof(*argv);
        argc--; argv++; plane_coefficients[3] = atof(*argv);
        R3Plane *plane = new R3Plane(plane_coefficients);
        planes.Insert(plane);
      }
      else if (!strcmp(*argv, "-camera")) { 
        RNCoord x, y, z, tx, ty, tz, ux, uy, uz;
        argv++; argc--; x = atof(*argv); 
        argv++; argc--; y = atof(*argv); 
        argv++; argc--; z = atof(*argv); 
        argv++; argc--; tx = atof(*argv); 
        argv++; argc--; ty = atof(*argv); 
        argv++; argc--; tz = atof(*argv); 
        argv++; argc--; ux = atof(*argv); 
        argv++; argc--; uy = atof(*argv); 
        argv++; argc--; uz = atof(*argv); 
        initial_camera_origin = R3Point(x, y, z);
        initial_camera_towards.Reset(tx, ty, tz);
        initial_camera_up.Reset(ux, uy, uz);
        initial_camera = TRUE;
      }
      else { fprintf(stderr, "Invalid program argument: %s", *argv); exit(1); }
      argv++; argc--;
    }
    else {
      if (!input_mesh_name) input_mesh_name = *argv;
      else if (!output_mesh_name) output_mesh_name = *argv;
      else { fprintf(stderr, "Invalid program argument: %s", *argv); exit(1); }
      argv++; argc--;
    }
  }

  // Check input filename
  if (!input_mesh_name) {
    fprintf(stderr, "You did not specify an input file name.\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



////////////////////////////////////////////////////////////////////////
// MAIN
////////////////////////////////////////////////////////////////////////

int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read mesh
  mesh = ReadMesh(input_mesh_name);
  if (!mesh) exit(-1);

  // Create viewer
  viewer = CreateBirdsEyeViewer(mesh);
  if (!viewer) exit(-1);

  // Check for correspondences
  if (input_correspondences_name) {
    // Read vertex correspondences
    vertex_correspondences = ReadVertexCorrespondences(mesh, input_correspondences_name);
    if (!vertex_correspondences) exit(-1);

    // Create edge correspondences
    edge_correspondences = CreateEdgeCorrespondences(mesh, *vertex_correspondences);
    if (!edge_correspondences) exit(-1);

    // Simplify mesh
    if (!SimplifyMesh(mesh, nfaces, *vertex_correspondences, *edge_correspondences)) exit(-1);

    // Create face correspondences
    face_correspondences = CreateFaceCorrespondences(mesh, *vertex_correspondences, *edge_correspondences);
    if (!face_correspondences) exit(-1);
  }
  else {
    // Simplify mesh
    if (!SimplifyMesh(mesh, nfaces)) exit(-1);
  }

  // Write correspondences
  if (output_correspondences_name && input_correspondences_name) {
    if (!WriteCorrespondences(mesh, 
      *vertex_correspondences, *edge_correspondences, *face_correspondences, 
      output_correspondences_name)) exit(-1);
  }

  // Write mesh
  if (output_mesh_name) {
    if (!WriteMesh(mesh, output_mesh_name)) exit(-1);
  }

  // Run interactive program
  if (output_image_name || interactive) {
    GLUIInit(&argc, argv);
    GLUIMainLoop();
  }

  // Return success 
  return 0;
}

















