// Source file for the mesh viewer program



// Ringlet include files

#include "R3Shapes/R3Shapes.h"
#include "pointset.h"



// Program variables

static char *mesh_name = NULL;
static char *grid_name = NULL;
static int maxpoints = 1024;
static int grid_resolution[3] = { 64, 64, 64 };
static int sample_interior_only = 0;
static int weight_samples_by_angle = 0;
static int plane_symmetry = 0;
static int print_verbose = 0;



// Application variables

static R3Mesh *mesh = NULL;
static PointSet *points = NULL;
static R3Grid *grid = NULL;



static R3Mesh *
ReadMesh(char *mesh_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate mesh
  R3Mesh *mesh = new R3Mesh();
  assert(mesh);

  // Read mesh from file
  if (!mesh->ReadFile(mesh_name)) {
    delete mesh;
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read mesh ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Faces = %d\n", mesh->NFaces());
    printf("  # Edges = %d\n", mesh->NEdges());
    printf("  # Vertices = %d\n", mesh->NVertices());
    fflush(stdout);
  }

  // Return success
  return mesh;
}



static PointSet *
CreatePoints(R3Mesh *mesh, int maxpoints)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate point set
  PointSet *points = new PointSet();
  assert(points);

  // Compute area of faces
  RNArea total_area = 0.0;
  RNArea *areas = new RNArea [ mesh->NFaces() ];
  for (int i = 0; i < mesh->NFaces(); i++) {
    R3MeshFace *face = mesh->Face(i);
    areas[i] = mesh->FaceArea(face);
    total_area += areas[i];
  }
    
  // Generate random points
  RNSeedRandomScalar();
  for (int i = 0; i < mesh->NFaces(); i++) {
    // Get face
    R3MeshFace *face = mesh->Face(i);

    // Get face normal
    const R3Vector& normal = mesh->FaceNormal(face);

    // Determine number of points for face
    RNScalar ideal_face_npoints = maxpoints * areas[i] / total_area;
    int face_npoints = (int) ideal_face_npoints;
    RNScalar remainder = ideal_face_npoints - face_npoints;
    if (remainder > RNRandomScalar()) face_npoints++;

    // Generate random points in triangle
    for (int j = 0; j < face_npoints; j++) {
      if (points->NPoints() == maxpoints) break;
      RNScalar r1 = sqrt(RNRandomScalar());
      RNScalar r2 = RNRandomScalar();
      const R3Point& p0 = mesh->VertexPosition(mesh->VertexOnFace(face, 0));
      const R3Point& p1 = mesh->VertexPosition(mesh->VertexOnFace(face, 1));
      const R3Point& p2 = mesh->VertexPosition(mesh->VertexOnFace(face, 2));
      R3Point position = p0 * (1.0 - r1) + p1 * r1 * (1.0 - r2) + p2 * r1 * r2;

      // Insert point
      Point p;
      p.px = position.X();
      p.py = position.Y();
      p.pz = position.Z();
      p.nx = normal.X();
      p.ny = normal.Y();
      p.nz = normal.Z();
      points->Insert(p);
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Created points ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Points = %d\n", points->NPoints());
    fflush(stdout);
  }

  // Return point set
  return points;
}



static R3Grid *
CreateGrid(R3Mesh *mesh, PointSet *points)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Compute mesh center of mass
  RNScalar mesh_area = 0.0;
  R3Point mesh_centroid = R3zero_point;
  for (int i = 0; i < mesh->NFaces(); i++) {
    R3MeshFace *face = mesh->Face(i);
    RNArea area = mesh->FaceArea(face);
    mesh_centroid += area * mesh->FaceCentroid(face);
    mesh_area += area;
  }
  mesh_centroid /= mesh_area;

  // Compute average distance to center of mass
  RNScalar avg_distance = 0.0;
  for (int i = 0; i < mesh->NFaces(); i++) {
    R3MeshFace *face = mesh->Face(i);
    RNArea area = mesh->FaceArea(face);
    avg_distance += area * R3Distance(mesh->FaceCentroid(face), mesh_centroid);
  }
  avg_distance /= mesh_area;

  // Compute grid center
  int res = grid_resolution[0];
  if (res < grid_resolution[1]) res = grid_resolution[1];
  if (res < grid_resolution[2]) res = grid_resolution[2];
  R3Vector grid_diagonal(grid_resolution[0]-1, grid_resolution[1]-1, grid_resolution[2]-1);
  R3Vector grid_centroid = 0.5 * grid_diagonal;

  // Compute scale factor
  RNScalar scale = (0.5 * (res-1)) / (2.0 * avg_distance);

  // Set affine transformation
  R3Affine world_to_grid = R3identity_affine;
  world_to_grid.Translate(grid_centroid);
  world_to_grid.Scale(scale);
  world_to_grid.Translate(-(mesh_centroid.Vector()));
  grid->SetWorldToGridTransformation(world_to_grid);

  // Fill grid
  int nsamples = 0;
  for (int i = 0; i < points->NPoints(); i++) {
    Point& p1 = (*points)[i];
    R3Point point1(p1.px, p1.py, p1.pz);
    R3Vector normal1(p1.nx, p1.ny, p1.nz);
    for (int j = i+1; j < points->NPoints(); j++) {
      Point& p2 = (*points)[j];
      R3Point point2(p2.px, p2.py, p2.pz);
      R3Vector normal2(p2.nx, p2.ny, p2.nz);

      // Compute vector point2-point1
      R3Span span(point1, point2);
      
      // Compute sample value (weighted by distance)
      double value = 1;
      double normalized_distance = span.Length() / avg_distance;
      if (RNIsZero(normalized_distance)) continue;
      value /= normalized_distance * normalized_distance;

      // Weight sample by angles
      if (weight_samples_by_angle) {
        // Take dot product of span with normal at p1
        double dot1 = span.Vector().Dot(normal1);
        if ((sample_interior_only) && (dot1 >= 0.0)) continue;
        double dot1_weight = fabs(dot1);

        // Take dot product of vector with normal at p2
        double dot2 = span.Vector().Dot(normal2);
        if ((sample_interior_only) && (dot2 <= 0.0)) continue;
        double dot2_weight = fabs(dot2);

        // Adjust value based on angle weights
        value *= dot1_weight * dot2_weight;
      }

      // Render sample(s)
      if (plane_symmetry) {
        // Render samples along bisecting plane
        R3Point midpoint(0.5 * (point2 + point1));
        R3Vector normal(point2 - point1);
        normal.Normalize();
        R3Plane midplane(midpoint, normal);
        grid->RasterizeWorldPlane(midplane, value);
      }
      else {
        // Render sample at midpoint
        R3Point midpoint(0.5 * (point2 + point1));
        grid->RasterizeWorldPoint(midpoint, value);
      }

      // Increment number of samples
      nsamples++;
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Created grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Samples = %d\n", nsamples); 
    printf("  # Voxels = %d ( %d x %d x %d )\n", res * res * res, res, res, res);
    fflush(stdout);
  }

  // Return grid
  return grid;
}



int WriteGrid(R3Grid *grid, char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write the grid
  if (!grid->WriteFile(filename)) {
    fprintf(stderr, "Unable to write grid file %s\n", filename);
    return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("Wrote grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if ((argc == 2) && (*argv[1] == '-')) {
    printf("Usage: symsam [options]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { print_verbose = 1; }
      else if (!strcmp(*argv, "-npoints")) { argc--; argv++; maxpoints = atoi(*argv); }
      else if (!strcmp(*argv, "-weight_samples_by_angle")) { weight_samples_by_angle = 1; }
      else if (!strcmp(*argv, "-sample_interior_only")) { sample_interior_only = 1; }
      else if (!strcmp(*argv, "-plane_symmetry")) { plane_symmetry = 1; }
      else if (!strcmp(*argv, "-resolution")) { 
        argc--; argv++; grid_resolution[0] = atoi(*argv); 
        argc--; argv++; grid_resolution[1] = atoi(*argv); 
        argc--; argv++; grid_resolution[2] = atoi(*argv); 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
    else {
      if (!mesh_name) mesh_name = *argv;
      else if (!grid_name) grid_name = *argv;
      else { fprintf(stderr, "Invalid program argument: %s", *argv); exit(1); }
      argv++; argc--;
    }
  }

  // Check mesh filename
  if (!mesh_name) {
    fprintf(stderr, "You did not specify a mesh file name.\n");
    return 0;
  }

  // Check grid filename
  if (!grid_name) {
    fprintf(stderr, "You did not specify a grid file name.\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



int main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read mesh
  mesh = ReadMesh(mesh_name);
  if (!mesh) exit(-1);

  // Create points
  points = CreatePoints(mesh, maxpoints);
  if (!points) exit(-1);

  // Create grid
  grid = CreateGrid(mesh, points);
  if (!grid) exit(-1);

  // Write grid
  if (!WriteGrid(grid, grid_name)) exit(-1);

  // Return success 
  return 0;
}

















